from typing import List, Dict

import os
import nltk
import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset
from transformers import AutoTokenizer
from transformers import DataCollatorForSeq2Seq

from SciAssist import BASE_CACHE_DIR
from SciAssist.datamodules.components.cora_label import label2id as cora_label2id
from SciAssist.models.components.bart_summarization import BartForSummarization
from SciAssist.models.components.flant5_summarization import FlanT5ForSummarization


class MyDatasetExtraction(Dataset):
    def __init__(self, tokenizer, dataset, token_pad_idx = 0, tag_pad_idx = -1):
        self.batch_size = 32
        self.max_len = 128
        self.token_pad_idx = token_pad_idx
        self.tag_pad_idx = tag_pad_idx
        self.tag2idx = {'B-DATA': 0, 'I-DATA': 1, 'O': 2}
        self.idx2tag = {0: 'B-DATA', 1: 'I-DATA', 2: 'O'}
        self.tokenizer = tokenizer
        self.dataset = self.preprocess(dataset)

     
    def __len__(self):
        """get dataset size"""
        return self.dataset['size']
 

    def __getitem__(self, idx):
        """sample data to get batch"""
        sentences = self.dataset['data'][idx]

        if 'labels' in self.dataset.keys():
            labels = self.dataset['labels'][idx]
            tags = self.dataset['tags'][idx]
            return [sentences, labels, tags]
        else:
            return [sentences]


    def preprocess(self, dataset):
        """Loads sentences and tags from their corresponding files. 
            Maps tokens and tags to their indices and stores them in the provided dict d.
        """
        processed_dataset = {}
        sentences = []
        if 'labels' in dataset.keys():
            labels = []
            tags = []

        for line in dataset['data']['text']:
            # replace each token by its index
            tokens = line.strip().split(' ')
            # print(tokens)
            subwords = list(map(self.tokenizer.tokenize, tokens)) # 每个词切分成子词
            # print(subwords)
            subword_lengths = list(map(len, subwords)) # 记录子词的长度，用于对齐tag
            # print(subword_lengths)
            subwords = ['[CLS]'] + [item for indices in subwords for item in indices]
            # subwords = ['<s>'] + [item for indices in subwords for item in indices] # 组成输入 token
            # print(subwords)
            token_start_idxs = 1 + np.cumsum([0] + subword_lengths[:-1]) # 记录每个token开始的位置
            # print(token_start_idxs)
            # print(self.tokenizer.convert_tokens_to_ids(subwords), token_start_idxs)
            sentences.append((self.tokenizer.convert_tokens_to_ids(subwords), token_start_idxs))

        if 'labels' in dataset.keys():
            for line in dataset['labels']['text']:
                labels.append(int(line.strip()))
            assert len(sentences) == len(labels)
            processed_dataset['labels'] = labels

            for line in dataset['tags']['text']:
                # replace each tag by its index
                tag_seq = [self.tag2idx.get(tag) for tag in line.strip().split(' ')]
                tags.append(tag_seq)
            # checks to ensure there is a tag for each token
            assert len(sentences) == len(tags)
            for i in range(len(sentences)):
                assert len(tags[i]) == len(sentences[i][-1])
            processed_dataset['tags'] = tags

        processed_dataset['data'] = sentences
        processed_dataset['size'] = len(sentences)

        return processed_dataset


    def collate_fn(self, batch):
        sentences = [x[0] for x in batch]
        processed_batch = {}
        if len(batch[0]) == 3:
            labels = [x[1] for x in batch]
            tags = [x[2] for x in batch]

        # batch length
        batch_len = len(sentences)  # batch size
        batch_max_subwords_len = max([len(s[0]) for s in sentences])
        max_subword_len = min(batch_max_subwords_len, self.max_len)
        max_token_len = 0
 
        # padding data 初始化
        batch_data = self.token_pad_idx * np.ones((batch_len, max_subword_len))
        batch_token_starts = []
 
        # padding and aligning
        for j in range(batch_len):
            cur_subwords_len = len(sentences[j][0])  # word_id list
            if cur_subwords_len <= max_subword_len:
                batch_data[j][:cur_subwords_len] = sentences[j][0]
            else:
                batch_data[j] = sentences[j][0][:max_subword_len]
            token_start_ids = sentences[j][-1]
            token_starts = np.zeros(max_subword_len)
            token_starts[[idx for idx in token_start_ids if idx < max_subword_len]] = 1
            batch_token_starts.append(token_starts)
            max_token_len = max(int(sum(token_starts)), max_token_len)

        processed_batch['input_subwords'] = torch.tensor(batch_data, dtype = torch.long)
        processed_batch['input_token_start_indexs'] = torch.tensor(np.array(batch_token_starts), dtype = torch.long)
        # processed_batch['attention_mask'] = (processed_batch['input_subwords'] != 1)
        processed_batch['attention_mask'] = processed_batch['input_subwords'].gt(0)

        if len(batch[0]) == 3:
            batch_tags = self.tag_pad_idx * np.ones((batch_len, max_token_len))
            batch_labels = np.ones((batch_len, ))
            for j in range(batch_len):
                batch_labels[j] = labels[j]
                cur_tags_len = len(tags[j])
                if cur_tags_len <= max_token_len:
                    batch_tags[j][:cur_tags_len] = tags[j]
                else:
                    batch_tags[j] = tags[j][:max_token_len]
            processed_batch['ner_tags'] = torch.tensor(np.array(batch_tags), dtype = torch.long)
            processed_batch['cls_labels'] = torch.tensor(np.array(batch_labels), dtype = torch.long)

        return processed_batch


class DataUtilsForDatasetExtraction():
    """

    Args:
        tokenizer (`PretrainedTokenizer`, default to None):
            The tokenizer for tokenization.
        checkpoint (`str`):
            The checkpoint from which the tokenizer is loaded.
        model_max_length (`int`, *optional*): The max sequence length the model accepts.
        max_source_length (`int`, *optional*): The max length of the input text.
        max_target_length (`int`, *optional*): The max length of the generated summary.
    """
    def __init__(self, tokenizer = None,
                 checkpoint = "allenai/scibert_scivocab_uncased",
                 model_max_length = 128
                 ):

        self.checkpoint = checkpoint
        self.model_max_length = model_max_length

        if tokenizer is None:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.checkpoint,
                model_max_length = self.model_max_length,
                cache_dir=BASE_CACHE_DIR,
                use_fast=True
            )
        else:
            self.tokenizer = tokenizer

        self.tag2idx = {'B-DATA': 0, 'I-DATA': 1, 'O': 2}
        self.idx2tag = {0: 'B-DATA', 1: 'I-DATA', 2: 'O'}


    def tokenize_and_align_labels(self, dataset):

        """

        Process the dataset for model input, for example, do tokenization and prepare label_ids.

        Args:
            dataset (`Dataset`): { "text": [s1, s2, ...], "summary": [l1, l2, ...]}
            inputs (`str`): The name of input column
            labels (`str`): The name of target column

        Returns:
            `Dict`: {"input_ids": input_ids, "attention_mask": attention_mask, "labels": label_ids }

        """
        
        """Loads sentences and tags from their corresponding files.
            Maps tokens and tags to their indices and stores them in the provided dict d.
        """
        processed_dataset = MyDatasetExtraction(self.tokenizer, dataset)

        return processed_dataset


    def postprocess(self, ner_output, cls_output, batch_tags, batch_labels):

        """
        Process model's outputs and get the final results rather than simple ids.

        Args:
            preds (Tensor): Prediction labels, the output of the model.
            labels (Tensor): True labels

        Returns:
            `(LongTensor, LongTensor)`: decoded_preds, decoded_labels

        """
        pred_tags = []
        true_tags = []
        pred_labels = []
        true_labels = []

        ner_output = ner_output.detach().cpu().numpy()
        cls_output = cls_output.detach().cpu().numpy()
        batch_tags = batch_tags.to('cpu').numpy()
        batch_labels = batch_labels.to('cpu').numpy()

        pred_tags.extend([[self.idx2tag.get(idx) for idx in indices] for indices in np.argmax(ner_output, axis=2)])
        true_tags.extend([[self.idx2tag.get(idx) if idx != -1 else 'O' for idx in indices] for indices in batch_tags])
        true_labels.extend(batch_labels)

        pred_labels.extend(cls_output)
        pred_labels = np.argmax(pred_labels, axis=1)

        assert len(pred_tags) == len(true_tags)

        return pred_tags, true_tags, pred_labels, true_labels


    def get_dataloader(self, dataset):

        """
        Generate DataLoader for a dataset.

        Args:
            dataset (`Dataset`): The raw dataset.
            inputs_column (`str`): Column name of the inputs.
            labels_column (`str`): Column name of the labels.

        Returns:
            `DataLoader`: A dataloader for the dataset. Will be used for inference.
        """
        tokenized_dataset = self.tokenize_and_align_labels(dataset)

        dataloader = DataLoader(
            dataset=tokenized_dataset,
            batch_size=32,
            collate_fn=tokenized_dataset.collate_fn,
        )

        return dataloader


class DataUtilsForSeq2Seq():
    """

    Args:
        tokenizer (`PretrainedTokenizer`, default to None):
            The tokenizer for tokenization.
        checkpoint (`str`):
            The checkpoint from which the tokenizer is loaded.
        model_max_length (`int`, *optional*): The max sequence length the model accepts.
        max_source_length (`int`, *optional*): The max length of the input text.
        max_target_length (`int`, *optional*): The max length of the generated summary.
    """


    def __init__(self, tokenizer = None, model_class = BartForSummarization,
                 checkpoint = "facebook/bart-large-cnn",
                 model_max_length = 1024,
                 max_source_length = 1024,
                 max_target_length = 128,
                 ):

        self.checkpoint = checkpoint
        self.model_max_length = model_max_length
        self.max_source_length = max_source_length
        self.max_target_length = max_target_length
        self.model_class = model_class

        if tokenizer is None:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.checkpoint,
                model_max_length = self.model_max_length,
                cache_dir=BASE_CACHE_DIR,
                use_fast=True
            )
        else:
            self.tokenizer = tokenizer


    def tokenize_and_align_labels(self, examples, inputs_column="text", labels_column="summary"):

        """

        Process the dataset for model input, for example, do tokenization and prepare label_ids.

        Args:
            examples (`Dataset`): { "text": [s1, s2, ...], "summary": [l1, l2, ...]}
            inputs (`str`): The name of input column
            labels (`str`): The name of target column

        Returns:
            `Dict`: {"input_ids": input_ids, "attention_mask": attention_mask, "labels": label_ids }

        """

        # Select input column
        inputs = examples[inputs_column]

        # Setup the tokenizer for inputs
        model_inputs = self.tokenizer(inputs, max_length=self.max_source_length, padding="max_length", truncation=True)

        # Select target column
        if labels_column in examples.keys():
            labels = examples[labels_column]
            # Setup the tokenizer for targets
            with self.tokenizer.as_target_tokenizer():
                labels = self.tokenizer(labels, max_length=self.max_target_length, padding="max_length", truncation=True)
                # Ignore padding in the loss
                labels["input_ids"] = [
                    [(l if l != self.tokenizer.pad_token_id else -100) for l in label] for label in labels["input_ids"]
                ]
            model_inputs["labels"] = labels["input_ids"]

        return model_inputs

    def collator(self):

        """

        The collating function.

        Returns:
            `function`: A collating function.

            For example, **DataCollatorForSeq2Seq(...)**.

            You can also custom a collating function, but remember that `collator()` needs to return a **function**.
        """


        return DataCollatorForSeq2Seq(self.tokenizer, model=self.model_class, pad_to_multiple_of=8)

    def postprocess(self, preds, labels):

        """
        Process model's outputs and get the final results rather than simple ids.

        Args:
            preds (Tensor): Prediction labels, the output of the model.
            labels (Tensor): True labels

        Returns:
            `(LongTensor, LongTensor)`: decoded_preds, decoded_labels

        """

        decoded_preds = self.tokenizer.batch_decode(preds, skip_special_tokens=True)

        labels = np.array(labels.to("cpu"))
        # Replace -100 in the labels as we can't decode them.
        labels = np.where(labels != -100, labels, self.tokenizer.pad_token_id)

        decoded_labels = self.tokenizer.batch_decode(labels, skip_special_tokens=True)

        decoded_preds = [pred.strip() for pred in decoded_preds]
        decoded_labels = [label.strip() for label in decoded_labels]

        # rougeLSum expects newline after each sentence
        decoded_preds = ["\n".join(nltk.sent_tokenize(pred)) for pred in decoded_preds]
        decoded_labels = ["\n".join(nltk.sent_tokenize(label)) for label in decoded_labels]

        return decoded_preds, decoded_labels

    def get_dataloader(self, dataset, inputs_column="text", labels_column="summary"):

        """
        Generate DataLoader for a dataset.

        Args:
            dataset (`Dataset`): The raw dataset.
            inputs_column (`str`): Column name of the inputs.
            labels_column (`str`): Column name of the labels.

        Returns:
            `DataLoader`: A dataloader for the dataset. Will be used for inference.
        """

        tokenized_example = dataset.map(
            lambda x: self.tokenize_and_align_labels(x, inputs_column=inputs_column, labels_column=labels_column),
            batched=True,
            remove_columns=dataset.column_names
        )

        dataloader = DataLoader(
            dataset=tokenized_example,
            batch_size=8,
            collate_fn=self.collator(),
        )

        return dataloader


class DataUtilsForTokenClassification():

    def __init__(self, tokenizer = None,
                 checkpoint ="allenai/scibert_scivocab_uncased",
                 model_max_length = 512,
                 ):

        self.checkpoint = checkpoint
        self.model_max_length = model_max_length

        if tokenizer == None:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.checkpoint,
                model_max_length=self.model_max_length,
                cache_dir=BASE_CACHE_DIR
            )
        else:
            self.tokenizer = tokenizer

    def tokenize_and_align_labels(self, examples, label2id=None):
        '''

        Prepare the dataset for input.
        For token-level task, construct token_mapping to obtain token based BERT representation from subtoken based one.

        Args:
            examples: Dataset, {"tokens":[[s1],[s2]..],"labels":[[l1],[l2]..]}
            label2id: Map label to label_id

        Returns:
            Dict{
                "input_ids":,
                "token_type_ids":,
                "attention_mask":,
                "token_mapping":,
                "labels":,
            }
        '''

        # Get input_ids, token_type_ids, attention_mask
        tokenized_inputs = self.tokenizer(
            examples["tokens"], truncation=True, is_split_into_words=True
        )
        if "labels" in examples.keys():
            raw_labels = examples["labels"]
            # label2id
            labels = [[int(label2id[l]) for l in label] for label in raw_labels]
            tokenized_inputs["labels"] = labels
            # Map sub-token to token
        tokenized_inputs["word_ids"] = []
        for i in range(len(examples["tokens"])):
            tokenized_inputs["word_ids"].append(tokenized_inputs.word_ids(i))

        # Prepare token_mapping for obtaining token based BERT representeation
        # Construct a subtoken to token mapping matrix token_mapping mapping [bsize, max_tok_len, max_subtok_len].
        # For example, in sent i, token j include subtokens[s:t), then mapping[i, j, s:t] = 1 / (t - s)
        # after obtaining subtoken based BERT representation `subtoken_context`[bsize, max_subtok_len, 768], use torch.matmul()
        # to obtain token based BERT representation
        # token_context = torch.matmul(token_mapping, subtoken_context)
        token_mappings = []
        for tokens, word_ids in zip(examples["tokens"], tokenized_inputs["word_ids"]):
            current_tok = 0

            # len(subtok_count) == the length of tokens for input, maybe smaller than origin ones
            # calculate the number of subtokens of a token
            subtok_count = [0]
            for tok_id in word_ids:
                if tok_id == None:
                    continue
                if tok_id == current_tok:
                    subtok_count[current_tok] += 1
                else:
                    current_tok += 1
                    subtok_count.append(1)
            # construct token_mapping
            token_mapping = []
            for i in range(len(subtok_count)):
                token_mapping.append([])
                for j in range(len(word_ids)):
                    token_mapping[i].append(0)

            for subtok_id, tok_id in enumerate(word_ids):
                if tok_id == None:
                    continue
                token_mapping[tok_id][subtok_id] = 1 / subtok_count[tok_id]

            token_mappings.append(token_mapping)

        tokenized_inputs["token_mapping"] = token_mappings
        return tokenized_inputs

    def convert_to_list(self, batch):
        res = []
        for i in batch:
            input_ids = i["input_ids"]
            token_type_ids = i["token_type_ids"]
            attn_mask = i["attention_mask"]
            token_mapping = i["token_mapping"]
            if "labels" in i.keys():
                labels = i["labels"]
                res.append([input_ids, token_type_ids, attn_mask, token_mapping, labels])
            else:
                res.append([input_ids, token_type_ids, attn_mask, token_mapping])
        return res

    def pad(self, batch: List[Dict]):
        # Pads to the longest sample
        batch = self.convert_to_list(batch)
        get_element = lambda x: [sample[x] for sample in batch]
        # subtoken length
        subtok_len = [len(tokens) for tokens in get_element(0)]
        max_subtok_len = np.array(subtok_len).max()
        # origin token length
        tok_len = [len(tokens) for tokens in get_element(3)]
        max_tok_len = np.array(tok_len).max()

        do_pad = lambda x, seqlen: [sample[x] + [0] * (seqlen - len(sample[x])) for sample in batch]  # 0: <pad>
        do_labels_pad = lambda x, seqlen: [sample[x] + [-100] * (seqlen - len(sample[x])) for sample in batch]

        # pad for origin tokens
        do_map_pad1 = lambda x, seqlen: [sample[x] + [[0]] * (seqlen - len(sample[x])) for sample in batch]
        # pad for subtokens
        do_map_pad2 = lambda batch, seqlen: [[subtoks + [0] * (seqlen - len(subtoks)) for subtoks in sample] for sample in batch]

        input_ids = do_pad(0, max_subtok_len)
        token_type_ids = do_pad(1, max_subtok_len)
        attn_mask = do_pad(2, max_subtok_len)
        token_mapping = do_map_pad1(3, max_tok_len)
        token_mapping = do_map_pad2(token_mapping, max_subtok_len)  # [batch_size, max_tok_len, max_subtok_len]

        LT = torch.LongTensor

        input_ids = LT(input_ids)
        attn_mask = LT(attn_mask)
        token_type_ids = LT(token_type_ids)
        token_mapping = torch.Tensor(token_mapping)
        if len(batch[0]) == 5:
            labels = do_labels_pad(4, max_tok_len)
            labels = LT(labels)
        else:
            return {
                "input_ids": input_ids,
                "token_type_ids": token_type_ids,
                "attention_mask": attn_mask,
                "token_mapping": token_mapping
            }

        return {
            "input_ids": input_ids,
            "token_type_ids": token_type_ids,
            "attention_mask": attn_mask,
            "labels": labels,
            "token_mapping": token_mapping
        }

    def collator(self):
        return self.pad

    def postprocess(self, preds, labels, label_names):
        '''

        Remove `-100` label and mask the padded labels with len(label_names).
        Args:
            preds (Tensor): Prediction labels
            labels (Tensor): True labels
            label_names (List): Label types

        Returns:
            (LongTensor, LongTensor):

        '''

        preds = preds.tolist()
        labels = labels.tolist()
        do_pad = lambda x, seqlen: [x + [len(label_names)] * (seqlen - len(x))]
        true_preds, true_labels = [], []
        for pred, label in zip(preds, labels):
            true_len = 0
            for l in label:
                if l == -100:
                    break
                else:
                    true_len += 1

            true_preds.append(do_pad(pred[:true_len], len(label)))
            true_labels.append(do_pad(label[:true_len], len(label)))
        true_labels = torch.LongTensor(true_labels)
        true_preds = torch.LongTensor(true_preds)
        return true_preds, true_labels

    def get_dataloader(self, dataset, label2id = cora_label2id):

        tokenized_example = dataset.map(
            lambda x: self.tokenize_and_align_labels(x, label2id),
            batched=True,
            remove_columns=dataset.column_names
        )
        dataloader = DataLoader(
            dataset=tokenized_example,
            batch_size=8,
            collate_fn=self.collator(),
        )

        return dataloader


# class DataUtilsForFiD():
#     """
#
#     Args:
#         tokenizer (`PretrainedTokenizer`, default to None):
#             The tokenizer for tokenization.
#         checkpoint (`str`):
#             The checkpoint from which the tokenizer is loaded.
#         model_max_length (`int`, *optional*): The max sequence length the model accepts.
#         max_source_length (`int`, *optional*): The max length of the input text.
#         max_target_length (`int`, *optional*): The max length of the generated summary.
#     """
#
#     def __init__(self, tokenizer = None, model_class = FiDT5,
#                  checkpoint = "google/flan-t5-large",
#                  model_max_length = 64,
#                  max_source_length = 64,
#                  max_target_length = 128,
#                  ):
#
#         self.checkpoint = checkpoint
#         self.model_max_length = model_max_length
#         self.max_source_length = max_source_length
#         self.max_target_length = max_target_length
#         self.model_class = model_class
#
#         if tokenizer is None:
#             self.tokenizer = AutoTokenizer.from_pretrained(
#                 self.checkpoint,
#                 model_max_length = self.model_max_length,
#                 cache_dir=BASE_CACHE_DIR,
#             )
#         else:
#             self.tokenizer = tokenizer
#
#
#     def tokenize_and_align_labels(self, examples, inputs_column="text", labels_column="summary", token_per_paragraph=50):
#
#         """
#
#         Process the dataset for model input, for example, do tokenization and prepare label_ids.
#
#         Args:
#             examples (`Dataset`): { "text": [s1, s2, ...], "summary": [l1, l2, ...]}
#             inputs (`str`): The name of input column
#             labels (`str`): The name of target column
#
#         Returns:
#             `Dict`: {"input_ids": input_ids, "attention_mask": attention_mask, "labels": label_ids }
#
#         """
#
#         # Select input column
#         inputs = examples[inputs_column]
#         dataset = {"paragraphs": []}
#
#         for input in inputs:
#             texts = ["Please give a summary of the following text: "]
#             tokens = input.split(" ")
#             index = 0
#             while index+token_per_paragraph < min(len(tokens),120*token_per_paragraph):
#                 p = " ".join(tokens[index:index+token_per_paragraph])
#                 texts.append(p)
#                 index += token_per_paragraph
#             texts.append(" ".join(tokens[index:index+token_per_paragraph]))
#             dataset["paragraphs"].append(texts)
#
#
#         # Select target column
#         if labels_column in examples.keys():
#             labels = examples[labels_column]
#             dataset["labels"] = labels
#
#         return dataset
#
#     def collator(self):
#
#         """
#
#         The collating function.
#
#         Returns:
#             `function`: A collating function.
#
#             For example, **DataCollatorForSeq2Seq(...)**.
#
#             You can also custom a collating function, but remember that `collator()` needs to return a **function**.
#         """
#
#         from SciAssist.utils.collators.CollatorForFid import DataCollatorForFid
#
#         return DataCollatorForFid(self.max_source_length, self.tokenizer, self.max_target_length)
#
#     def postprocess(self, preds, labels):
#
#         """
#         Process model's outputs and get the final results rather than simple ids.
#
#         Args:
#             preds (Tensor): Prediction labels, the output of the model.
#             labels (Tensor): True labels
#
#         Returns:
#             `(LongTensor, LongTensor)`: decoded_preds, decoded_labels
#
#         """
#
#         decoded_preds = self.tokenizer.batch_decode(preds, skip_special_tokens=True)
#
#         labels = np.array(labels.to("cpu"))
#         # Replace -100 in the labels as we can't decode them.
#         labels = np.where(labels != -100, labels, self.tokenizer.pad_token_id)
#
#         decoded_labels = self.tokenizer.batch_decode(labels, skip_special_tokens=True)
#
#         decoded_preds = [pred.strip() for pred in decoded_preds]
#         decoded_labels = [label.strip() for label in decoded_labels]
#
#         # rougeLSum expects newline after each sentence
#         decoded_preds = ["\n".join(nltk.sent_tokenize(pred)) for pred in decoded_preds]
#         decoded_labels = ["\n".join(nltk.sent_tokenize(label)) for label in decoded_labels]
#
#         return decoded_preds, decoded_labels
#
#     def get_dataloader(self, dataset, inputs_column="text", labels_column="summary"):
#
#         """
#         Generate DataLoader for a dataset.
#
#         Args:
#             dataset (`Dataset`): The raw dataset.
#             inputs_column (`str`): Column name of the inputs.
#             labels_column (`str`): Column name of the labels.
#
#         Returns:
#             `DataLoader`: A dataloader for the dataset. Will be used for inference.
#         """
#
#         tokenized_example = dataset.map(
#             lambda x: self.tokenize_and_align_labels(x, inputs_column=inputs_column, labels_column=labels_column),
#             batched=True,
#             remove_columns=dataset.column_names
#         )
#         dataloader = DataLoader(
#             dataset=tokenized_example,
#             batch_size=8,
#             collate_fn=self.collator(),
#         )
#
#         return dataloader
#
# #
#
# class DataUtilsForFrost():
#     """
#
#     Args:
#         tokenizer (`PretrainedTokenizer`, default to None):
#             The tokenizer for tokenization.
#         checkpoint (`str`):
#             The checkpoint from which the tokenizer is loaded.
#         model_max_length (`int`, *optional*): The max sequence length the model accepts.
#         max_source_length (`int`, *optional*): The max length of the input text.
#         max_target_length (`int`, *optional*): The max length of the generated summary.
#     """
#
#
#     def __init__(self, tokenizer = None, model_class = FrostForSummarization,
#                  checkpoint = "pegasus/frost",
#                  model_max_length = 1024,
#                  max_source_length = 1024,
#                  max_target_length = 128,
#                  ):
#
#         self.checkpoint = checkpoint
#         self.model_max_length = model_max_length
#         self.max_source_length = max_source_length
#         self.max_target_length = max_target_length
#         self.model_class = model_class
#
#         if tokenizer is None:
#             self.tokenizer = PegasusTokenizer.from_pretrained(
#                 self.checkpoint,
#                 cache_dir=BASE_CACHE_DIR,
#                 model_max_length=self.model_max_length,
#             )
#         else:
#             self.tokenizer = tokenizer
#
#         # FROST Constants
#         self.ENTITYCHAIN_START_TOKEN = "[CONTENT]"
#         self.SUMMARY_START_TOKEN = "[SUMMARY]"
#         self.ENTITY_SEPARATOR = " | "
#         self.ENTITY_SENTENCE_SEPARATOR = " ||| "
#
#         # Prepare Spacy processor
#         self.SPACY_MODEL_OR_PATH = "en_core_web_sm"
#         self.SPACY_PROCESSOR = spacy.load(self.SPACY_MODEL_OR_PATH)
#
#     def get_frost_labels(self, text):
#         """Gets Spacy Frost processor."""
#         entity_plans = []
#         for text_sent in self.SPACY_PROCESSOR(text.replace("\n", " ")).sents:
#             entity_plans.append(
#                 self.ENTITY_SEPARATOR.join(
#                     [entity.text for entity in self.SPACY_PROCESSOR(text_sent.text).ents]))
#         text_with_entityplans = (
#                 self.ENTITYCHAIN_START_TOKEN + " " +
#                 self.ENTITY_SENTENCE_SEPARATOR.join(entity_plans) + " " +
#                 self.SUMMARY_START_TOKEN + " " + text)
#         return text_with_entityplans
#
#
#     def tokenize_and_align_labels(self, examples, inputs_column="text", labels_column="summary"):
#
#         """
#
#         Process the dataset for model input, for example, do tokenization and prepare label_ids.
#
#         Args:
#             examples (`Dataset`): { "text": [s1, s2, ...], "summary": [l1, l2, ...]}
#             inputs (`str`): The name of input column
#             labels (`str`): The name of target column
#
#         Returns:
#             `Dict`: {"input_ids": input_ids, "attention_mask": attention_mask, "labels": label_ids }
#
#         """
#
#         # Select input column
#         inputs = examples[inputs_column]
#
#         # Setup the tokenizer for inputs
#         model_inputs = self.tokenizer(inputs, max_length=self.max_target_length, padding="max_length", truncation=True)
#
#         # Select target column
#         if labels_column in examples.keys():
#             labels = examples[labels_column]
#             labels = [self.get_frost_labels(label) for label in labels]
#
#             # Setup the tokenizer for targets
#             with self.tokenizer.as_target_tokenizer():
#                 labels = self.tokenizer(labels, max_length=self.max_target_length, padding="max_length", truncation=True)
#                 # Ignore padding in the loss
#                 labels["input_ids"] = [
#                     [(l if l != self.tokenizer.pad_token_id else -100) for l in label] for label in labels["input_ids"]
#                 ]
#
#             model_inputs["labels"] = labels["input_ids"]
#
#         return model_inputs
#
#     def collator(self):
#
#         """
#
#         The collating function.
#
#         Returns:
#             `function`: A collating function.
#
#             For example, **DataCollatorForSeq2Seq(...)**.
#
#             You can also custom a collating function, but remember that `collator()` needs to return a **function**.
#         """
#
#
#         return DataCollatorForSeq2Seq(self.tokenizer, model=self.model_class, pad_to_multiple_of=8)
#
#     def postprocess(self, preds, labels):
#
#         """
#         Process model's outputs and get the final results rather than simple ids.
#
#         Args:
#             preds (Tensor): Prediction labels, the output of the model.
#             labels (Tensor): True labels
#
#         Returns:
#             `(LongTensor, LongTensor)`: decoded_preds, decoded_labels
#
#         """
#
#         decoded_preds = self.tokenizer.batch_decode(preds, skip_special_tokens=True)
#
#         labels = np.array(labels.to("cpu"))
#         # Replace -100 in the labels as we can't decode them.
#         labels = np.where(labels != -100, labels, self.tokenizer.pad_token_id)
#
#         decoded_labels = self.tokenizer.batch_decode(labels, skip_special_tokens=False)
#
#         decoded_preds = [pred.strip() for pred in decoded_preds]
#         decoded_labels = [label.strip() for label in decoded_labels]
#
#         # rougeLSum expects newline after each sentence
#         decoded_preds = ["\n".join(nltk.sent_tokenize(pred)) for pred in decoded_preds]
#         decoded_labels = ["\n".join(nltk.sent_tokenize(label)) for label in decoded_labels]
#
#         return decoded_preds, decoded_labels
#
#     def get_dataloader(self, dataset, inputs_column="text", labels_column="summary"):
#
#         """
#         Generate DataLoader for a dataset.
#
#         Args:
#             dataset (`Dataset`): The raw dataset.
#             inputs_column (`str`): Column name of the inputs.
#             labels_column (`str`): Column name of the labels.
#
#         Returns:
#             `DataLoader`: A dataloader for the dataset. Will be used for inference.
#         """
#
#         tokenized_example = dataset.map(
#             lambda x: self.tokenize_and_align_labels(x, inputs_column=inputs_column, labels_column=labels_column),
#             batched=True,
#             remove_columns=dataset.column_names
#         )
#         dataloader = DataLoader(
#             dataset=tokenized_example,
#             batch_size=8,
#             collate_fn=self.collator(),
#         )
#
#         return dataloader

class DataUtilsForT5():
    """

    Args:
        tokenizer (`PretrainedTokenizer`, default to None):
            The tokenizer for tokenization.
        checkpoint (`str`):
            The checkpoint from which the tokenizer is loaded.
        model_max_length (`int`, *optional*): The max sequence length the model accepts.
        max_source_length (`int`, *optional*): The max length of the input text.
        max_target_length (`int`, *optional*): The max length of the generated summary.
    """


    def __init__(self, tokenizer = None,
                 model_class = FlanT5ForSummarization,
                 checkpoint = "facebook/bart-large-cnn",
                 prompt = "Please give a summary of the following text: ",
                 model_max_length = 1024,
                 max_source_length = 1024,
                 max_target_length = 128,
                 ):

        self.checkpoint = checkpoint
        self.model_max_length = model_max_length
        self.max_source_length = max_source_length
        self.max_target_length = max_target_length
        self.model_class = model_class
        self.prompt = prompt

        if tokenizer is None:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.checkpoint,
                model_max_length = self.model_max_length,
                cache_dir=BASE_CACHE_DIR,
                use_fast=True
            )
        else:
            self.tokenizer = tokenizer


    def tokenize_and_align_labels(self, examples, inputs_column="text", labels_column="summary"):

        """

        Process the dataset for model input, for example, do tokenization and prepare label_ids.

        Args:
            examples (`Dataset`): { "text": [s1, s2, ...], "summary": [l1, l2, ...]}
            inputs (`str`): The name of input column
            labels (`str`): The name of target column

        Returns:
            `Dict`: {"input_ids": input_ids, "attention_mask": attention_mask, "labels": label_ids }

        """

        # Select input column
        inputs = examples[inputs_column]
        inputs = [self.prompt + raw_text for raw_text in inputs]
        # Setup the tokenizer for inputs
        model_inputs = self.tokenizer(inputs, max_length=self.max_source_length, padding="max_length", truncation=True)

        # Select target column
        if labels_column in examples.keys():
            labels = examples[labels_column]
            # Setup the tokenizer for targets
            with self.tokenizer.as_target_tokenizer():
                labels = self.tokenizer(labels, max_length=self.max_target_length, padding="max_length", truncation=True)
                # Ignore padding in the loss
                labels["input_ids"] = [
                    [(l if l != self.tokenizer.pad_token_id else -100) for l in label] for label in labels["input_ids"]
                ]
            model_inputs["labels"] = labels["input_ids"]

        return model_inputs

    def collator(self):

        """

        The collating function.

        Returns:
            `function`: A collating function.

            For example, **DataCollatorForSeq2Seq(...)**.

            You can also custom a collating function, but remember that `collator()` needs to return a **function**.
        """


        return DataCollatorForSeq2Seq(self.tokenizer, model=self.model_class, pad_to_multiple_of=8)

    def postprocess(self, preds, labels):

        """
        Process model's outputs and get the final results rather than simple ids.

        Args:
            preds (Tensor): Prediction labels, the output of the model.
            labels (Tensor): True labels

        Returns:
            `(LongTensor, LongTensor)`: decoded_preds, decoded_labels

        """

        decoded_preds = self.tokenizer.batch_decode(preds, skip_special_tokens=True)

        labels = np.array(labels.to("cpu"))
        # Replace -100 in the labels as we can't decode them.
        labels = np.where(labels != -100, labels, self.tokenizer.pad_token_id)

        decoded_labels = self.tokenizer.batch_decode(labels, skip_special_tokens=True)

        decoded_preds = [pred.strip() for pred in decoded_preds]
        decoded_labels = [label.strip() for label in decoded_labels]

        # rougeLSum expects newline after each sentence
        decoded_preds = ["\n".join(nltk.sent_tokenize(pred)) for pred in decoded_preds]
        decoded_labels = ["\n".join(nltk.sent_tokenize(label)) for label in decoded_labels]

        return decoded_preds, decoded_labels

    def get_dataloader(self, dataset, inputs_column="text", labels_column="summary"):

        """
        Generate DataLoader for a dataset.

        Args:
            dataset (`Dataset`): The raw dataset.
            inputs_column (`str`): Column name of the inputs.
            labels_column (`str`): Column name of the labels.

        Returns:
            `DataLoader`: A dataloader for the dataset. Will be used for inference.
        """

        tokenized_example = dataset.map(
            lambda x: self.tokenize_and_align_labels(x, inputs_column=inputs_column, labels_column=labels_column),
            batched=True,
            remove_columns=dataset.column_names
        )

        dataloader = DataLoader(
            dataset=tokenized_example,
            batch_size=8,
            collate_fn=self.collator(),
        )

        return dataloader


class DataUtilsForFlanT5():
    """

    Args:
        tokenizer (`PretrainedTokenizer`, default to None):
            The tokenizer for tokenization.
        checkpoint (`str`):
            The checkpoint from which the tokenizer is loaded.
        model_max_length (`int`, *optional*): The max sequence length the model accepts.
        max_source_length (`int`, *optional*): The max length of the input text.
        max_target_length (`int`, *optional*): The max length of the generated summary.
    """


    def __init__(self, tokenizer = None,
                 model_class = FlanT5ForSummarization,
                 checkpoint = "google/flan-t5-base",
                 prompt = "Please give a summary of the following text ",
                 model_max_length = 1024,
                 max_source_length = 1024,
                 max_target_length = 500,
                 ):

        self.checkpoint = checkpoint
        self.model_max_length = model_max_length
        self.max_source_length = max_source_length
        self.max_target_length = max_target_length
        self.model_class = model_class
        self.prompt = prompt



        if tokenizer is None:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.checkpoint,
                model_max_length = self.model_max_length,
                cache_dir=BASE_CACHE_DIR,
                use_fast=True
            )
        else:
            self.tokenizer = tokenizer


    def tokenize_and_align_labels(self, examples, inputs_column="text", labels_column="summary"):

        """

        Process the dataset for model input, for example, do tokenization and prepare label_ids.

        Args:
            examples (`Dataset`): { "text": [s1, s2, ...], "summary": [l1, l2, ...]}
            inputs (`str`): The name of input column
            labels (`str`): The name of target column

        Returns:
            `Dict`: {"input_ids": input_ids, "attention_mask": attention_mask, "labels": label_ids }

        """

        inputs = examples[inputs_column]
        prompts = [ self.prompt for i in inputs ]

        # kw_instructions = ["{}"]
        def kw(prompt, keyword):
            # i = randint(1,10)
            if keyword is not None and keyword != [""]:
                return "Keywords: [ " + str(", ".join(keyword)) + " ]. " + prompt + "based on these keywords " if keyword is not None else prompt
            return prompt
            # return prompt + "that focuses on " + str(", ".join(keyword)) + " " if keyword is not None else prompt
        def leng(prompt, length):
            if length is not None:
                return prompt + ", which has less than " + str(length) + " words " if length is not None else prompt
            return prompt

        # def k_str(k_list):
        #     if k_list is not None:
        #         return ", ".join(k_list)
        #     return None
        # def sents_num(l):
        #     if l is not None:
        #         return 5*math.ceil(l/50)
        #     return 10
        if "keywords" in examples.keys():
            keywords = examples["keywords"]
            if keywords is not None:
                prompts = [ kw(prompt,keyword) for (prompt,keyword) in zip(prompts,keywords) ]
                # keywords = [k_str(k) for k in keywords]
                # if "length" in examples.keys() and examples["length"] is not None:
                #     inputs = [extract_related_sentences(t,k,sents_num(l)) for (t,k,l) in zip(inputs,keywords,examples["length"])]
                # else:
                #     inputs = [extract_related_sentences(t, k, 10) for (t, k) in zip(inputs, keywords)]
        if "length" in examples.keys():
            if examples["length"] is not None:
                lengths = examples["length"]
                prompts = [ leng(prompt,length) for (prompt,length) in zip(prompts,lengths)]

        # kwords = ["" for i in inputs]
        # if "keywords" in examples.keys():
        #     keywords = examples["keywords"]
        #     if keywords is not None:
        #         keywords = [ keyword if keyword is not None else [] for keyword in keywords ]
        #         # print(keywords)
        #         kwords = [ " ".join(keyword) for keyword in keywords ]
        inputs = [ prompt + ": " + raw_text for (prompt,raw_text) in zip(prompts, inputs) ]
        # print(inputs[0][:200])


        # Setup the tokenizer for inputs

        model_inputs = self.tokenizer(inputs, max_length=self.max_source_length, padding="max_length", truncation=True)

        # Select target column
        if labels_column in examples.keys():
            labels = examples[labels_column]
            # Setup the tokenizer for targets
            with self.tokenizer.as_target_tokenizer():
                labels = self.tokenizer(labels, max_length=self.max_target_length, padding="max_length", truncation=True)
                # Ignore padding in the loss
                labels["input_ids"] = [
                    [(l if l != self.tokenizer.pad_token_id else -100) for l in label] for label in labels["input_ids"]
                ]
            model_inputs["labels"] = labels["input_ids"]

        #when test
        # if "id" in examples.keys():
        #     model_inputs["id"] = examples["id"]
        # if "length" in examples.keys() and examples["length"] is not None:
        #     if examples["length"][0] is not None:
        #         model_inputs["length"] = examples["length"]
        return model_inputs

    def collator(self):

        """

        The collating function.

        Returns:
            `function`: A collating function.

            For example, **DataCollatorForSeq2Seq(...)**.

            You can also custom a collating function, but remember that `collator()` needs to return a **function**.
        """


        return DataCollatorForSeq2Seq(self.tokenizer, model=self.model_class, pad_to_multiple_of=8)

    def postprocess(self, preds, labels):

        """
        Process model's outputs and get the final results rather than simple ids.

        Args:
            preds (Tensor): Prediction labels, the output of the model.
            labels (Tensor): True labels

        Returns:
            `(LongTensor, LongTensor)`: decoded_preds, decoded_labels

        """

        decoded_preds = self.tokenizer.batch_decode(preds, skip_special_tokens=True)

        labels = np.array(labels.to("cpu"))
        # Replace -100 in the labels as we can't decode them.
        labels = np.where(labels != -100, labels, self.tokenizer.pad_token_id)

        decoded_labels = self.tokenizer.batch_decode(labels, skip_special_tokens=True)

        decoded_preds = [pred.strip() for pred in decoded_preds]
        decoded_labels = [label.strip() for label in decoded_labels]

        # rougeLSum expects newline after each sentence
        decoded_preds = ["\n".join(nltk.sent_tokenize(pred)) for pred in decoded_preds]
        decoded_labels = ["\n".join(nltk.sent_tokenize(label)) for label in decoded_labels]

        return decoded_preds, decoded_labels

    def get_dataloader(self, dataset, inputs_column="text", labels_column="summary"):

        """
        Generate DataLoader for a dataset.

        Args:
            dataset (`Dataset`): The raw dataset.
            inputs_column (`str`): Column name of the inputs.
            labels_column (`str`): Column name of the labels.

        Returns:
            `DataLoader`: A dataloader for the dataset. Will be used for inference.
        """

        tokenized_example = dataset.map(
            lambda x: self.tokenize_and_align_labels(x, inputs_column=inputs_column, labels_column=labels_column),
            batched=True,
            remove_columns=dataset.column_names
        )

        dataloader = DataLoader(
            dataset=tokenized_example,
            batch_size=8,
            collate_fn=self.collator(),
        )

        return dataloader


# class DataUtilsForFiD():
#     """
#
#     Args:
#         tokenizer (`PretrainedTokenizer`, default to None):
#             The tokenizer for tokenization.
#         checkpoint (`str`):
#             The checkpoint from which the tokenizer is loaded.
#         model_max_length (`int`, *optional*): The max sequence length the model accepts.
#         max_source_length (`int`, *optional*): The max length of the input text.
#         max_target_length (`int`, *optional*): The max length of the generated summary.
#     """
#
#     def __init__(self, tokenizer = None, model_class = FiDT5,
#                  checkpoint = "google/flan-t5-large",
#                  model_max_length = 64,
#                  max_source_length = 64,
#                  max_target_length = 128,
#                  ):
#
#         self.checkpoint = checkpoint
#         self.model_max_length = model_max_length
#         self.max_source_length = max_source_length
#         self.max_target_length = max_target_length
#         self.model_class = model_class
#
#         if tokenizer is None:
#             self.tokenizer = AutoTokenizer.from_pretrained(
#                 self.checkpoint,
#                 model_max_length = self.model_max_length,
#                 cache_dir=BASE_CACHE_DIR,
#             )
#         else:
#             self.tokenizer = tokenizer
#
#
#     def tokenize_and_align_labels(self, examples, inputs_column="text", labels_column="summary", token_per_paragraph=50):
#
#         """
#
#         Process the dataset for model input, for example, do tokenization and prepare label_ids.
#
#         Args:
#             examples (`Dataset`): { "text": [s1, s2, ...], "summary": [l1, l2, ...]}
#             inputs (`str`): The name of input column
#             labels (`str`): The name of target column
#
#         Returns:
#             `Dict`: {"input_ids": input_ids, "attention_mask": attention_mask, "labels": label_ids }
#
#         """
#
#         # Select input column
#         inputs = examples[inputs_column]
#         dataset = {"paragraphs": []}
#
#         for input in inputs:
#             texts = ["Please give a summary of the following text: "]
#             tokens = input.split(" ")
#             index = 0
#             while index+token_per_paragraph < min(len(tokens),120*token_per_paragraph):
#                 p = " ".join(tokens[index:index+token_per_paragraph])
#                 texts.append(p)
#                 index += token_per_paragraph
#             texts.append(" ".join(tokens[index:index+token_per_paragraph]))
#             dataset["paragraphs"].append(texts)
#
#
#         # Select target column
#         if labels_column in examples.keys():
#             labels = examples[labels_column]
#             dataset["labels"] = labels
#
#         return dataset
#
#     def collator(self):
#
#         """
#
#         The collating function.
#
#         Returns:
#             `function`: A collating function.
#
#             For example, **DataCollatorForSeq2Seq(...)**.
#
#             You can also custom a collating function, but remember that `collator()` needs to return a **function**.
#         """
#
#         from SciAssist.utils.collators.CollatorForFid import DataCollatorForFid
#
#         return DataCollatorForFid(self.max_source_length, self.tokenizer, self.max_target_length)
#
#     def postprocess(self, preds, labels):
#
#         """
#         Process model's outputs and get the final results rather than simple ids.
#
#         Args:
#             preds (Tensor): Prediction labels, the output of the model.
#             labels (Tensor): True labels
#
#         Returns:
#             `(LongTensor, LongTensor)`: decoded_preds, decoded_labels
#
#         """
#
#         decoded_preds = self.tokenizer.batch_decode(preds, skip_special_tokens=True)
#
#         labels = np.array(labels.to("cpu"))
#         # Replace -100 in the labels as we can't decode them.
#         labels = np.where(labels != -100, labels, self.tokenizer.pad_token_id)
#
#         decoded_labels = self.tokenizer.batch_decode(labels, skip_special_tokens=True)
#
#         decoded_preds = [pred.strip() for pred in decoded_preds]
#         decoded_labels = [label.strip() for label in decoded_labels]
#
#         # rougeLSum expects newline after each sentence
#         decoded_preds = ["\n".join(nltk.sent_tokenize(pred)) for pred in decoded_preds]
#         decoded_labels = ["\n".join(nltk.sent_tokenize(label)) for label in decoded_labels]
#
#         return decoded_preds, decoded_labels
#
#     def get_dataloader(self, dataset, inputs_column="text", labels_column="summary"):
#
#         """
#         Generate DataLoader for a dataset.
#
#         Args:
#             dataset (`Dataset`): The raw dataset.
#             inputs_column (`str`): Column name of the inputs.
#             labels_column (`str`): Column name of the labels.
#
#         Returns:
#             `DataLoader`: A dataloader for the dataset. Will be used for inference.
#         """
#
#         tokenized_example = dataset.map(
#             lambda x: self.tokenize_and_align_labels(x, inputs_column=inputs_column, labels_column=labels_column),
#             batched=True,
#             remove_columns=dataset.column_names
#         )
#         dataloader = DataLoader(
#             dataset=tokenized_example,
#             batch_size=8,
#             collate_fn=self.collator(),
#         )
#
#         return dataloader
#
#
#
# class DataUtilsForFrost():
#     """
#
#     Args:
#         tokenizer (`PretrainedTokenizer`, default to None):
#             The tokenizer for tokenization.
#         checkpoint (`str`):
#             The checkpoint from which the tokenizer is loaded.
#         model_max_length (`int`, *optional*): The max sequence length the model accepts.
#         max_source_length (`int`, *optional*): The max length of the input text.
#         max_target_length (`int`, *optional*): The max length of the generated summary.
#     """
#
#
#     def __init__(self, tokenizer = None, model_class = FrostForSummarization,
#                  checkpoint = "pegasus/frost",
#                  model_max_length = 1024,
#                  max_source_length = 1024,
#                  max_target_length = 128,
#                  ):
#
#         self.checkpoint = checkpoint
#         self.model_max_length = model_max_length
#         self.max_source_length = max_source_length
#         self.max_target_length = max_target_length
#         self.model_class = model_class
#
#         if tokenizer is None:
#             self.tokenizer = PegasusTokenizer.from_pretrained(
#                 self.checkpoint,
#                 cache_dir=BASE_CACHE_DIR,
#                 model_max_length=self.model_max_length,
#             )
#         else:
#             self.tokenizer = tokenizer
#
#         # FROST Constants
#         self.ENTITYCHAIN_START_TOKEN = "[CONTENT]"
#         self.SUMMARY_START_TOKEN = "[SUMMARY]"
#         self.ENTITY_SEPARATOR = " | "
#         self.ENTITY_SENTENCE_SEPARATOR = " ||| "
#
#         # Prepare Spacy processor
#         self.SPACY_MODEL_OR_PATH = "en_core_web_sm"
#         self.SPACY_PROCESSOR = spacy.load(self.SPACY_MODEL_OR_PATH)
#
#     def get_frost_labels(self, text):
#         """Gets Spacy Frost processor."""
#         entity_plans = []
#         for text_sent in self.SPACY_PROCESSOR(text.replace("\n", " ")).sents:
#             entity_plans.append(
#                 self.ENTITY_SEPARATOR.join(
#                     [entity.text for entity in self.SPACY_PROCESSOR(text_sent.text).ents]))
#         text_with_entityplans = (
#                 self.ENTITYCHAIN_START_TOKEN + " " +
#                 self.ENTITY_SENTENCE_SEPARATOR.join(entity_plans) + " " +
#                 self.SUMMARY_START_TOKEN + " " + text)
#         return text_with_entityplans
#
#
#     def tokenize_and_align_labels(self, examples, inputs_column="text", labels_column="summary"):
#
#         """
#
#         Process the dataset for model input, for example, do tokenization and prepare label_ids.
#
#         Args:
#             examples (`Dataset`): { "text": [s1, s2, ...], "summary": [l1, l2, ...]}
#             inputs (`str`): The name of input column
#             labels (`str`): The name of target column
#
#         Returns:
#             `Dict`: {"input_ids": input_ids, "attention_mask": attention_mask, "labels": label_ids }
#
#         """
#
#         # Select input column
#         inputs = examples[inputs_column]
#
#         # Setup the tokenizer for inputs
#         model_inputs = self.tokenizer(inputs, max_length=self.max_target_length, padding="max_length", truncation=True)
#
#         # Select target column
#         if labels_column in examples.keys():
#             labels = examples[labels_column]
#             labels = [self.get_frost_labels(label) for label in labels]
#
#             # Setup the tokenizer for targets
#             with self.tokenizer.as_target_tokenizer():
#                 labels = self.tokenizer(labels, max_length=self.max_target_length, padding="max_length", truncation=True)
#                 # Ignore padding in the loss
#                 labels["input_ids"] = [
#                     [(l if l != self.tokenizer.pad_token_id else -100) for l in label] for label in labels["input_ids"]
#                 ]
#
#             model_inputs["labels"] = labels["input_ids"]
#
#         return model_inputs
#
#     def collator(self):
#
#         """
#
#         The collating function.
#
#         Returns:
#             `function`: A collating function.
#
#             For example, **DataCollatorForSeq2Seq(...)**.
#
#             You can also custom a collating function, but remember that `collator()` needs to return a **function**.
#         """
#
#
#         return DataCollatorForSeq2Seq(self.tokenizer, model=self.model_class, pad_to_multiple_of=8)
#
#     def postprocess(self, preds, labels):
#
#         """
#         Process model's outputs and get the final results rather than simple ids.
#
#         Args:
#             preds (Tensor): Prediction labels, the output of the model.
#             labels (Tensor): True labels
#
#         Returns:
#             `(LongTensor, LongTensor)`: decoded_preds, decoded_labels
#
#         """
#
#         decoded_preds = self.tokenizer.batch_decode(preds, skip_special_tokens=True)
#
#         labels = np.array(labels.to("cpu"))
#         # Replace -100 in the labels as we can't decode them.
#         labels = np.where(labels != -100, labels, self.tokenizer.pad_token_id)
#
#         decoded_labels = self.tokenizer.batch_decode(labels, skip_special_tokens=False)
#
#         decoded_preds = [pred.strip() for pred in decoded_preds]
#         decoded_labels = [label.strip() for label in decoded_labels]
#
#         # rougeLSum expects newline after each sentence
#         decoded_preds = ["\n".join(nltk.sent_tokenize(pred)) for pred in decoded_preds]
#         decoded_labels = ["\n".join(nltk.sent_tokenize(label)) for label in decoded_labels]
#
#         return decoded_preds, decoded_labels
#
#     def get_dataloader(self, dataset, inputs_column="text", labels_column="summary"):
#
#         """
#         Generate DataLoader for a dataset.
#
#         Args:
#             dataset (`Dataset`): The raw dataset.
#             inputs_column (`str`): Column name of the inputs.
#             labels_column (`str`): Column name of the labels.
#
#         Returns:
#             `DataLoader`: A dataloader for the dataset. Will be used for inference.
#         """
#
#         tokenized_example = dataset.map(
#             lambda x: self.tokenize_and_align_labels(x, inputs_column=inputs_column, labels_column=labels_column),
#             batched=True,
#             remove_columns=dataset.column_names
#         )
#         dataloader = DataLoader(
#             dataset=tokenized_example,
#             batch_size=8,
#             collate_fn=self.collator(),
#         )
#
#         return dataloader

class DataUtilsForT5():
    """

    Args:
        tokenizer (`PretrainedTokenizer`, default to None):
            The tokenizer for tokenization.
        checkpoint (`str`):
            The checkpoint from which the tokenizer is loaded.
        model_max_length (`int`, *optional*): The max sequence length the model accepts.
        max_source_length (`int`, *optional*): The max length of the input text.
        max_target_length (`int`, *optional*): The max length of the generated summary.
    """


    def __init__(self, tokenizer = None,
                 model_class = BartForSummarization,
                 checkpoint = "facebook/bart-large-cnn",
                 prompt = "Please give a summary of the following text: ",
                 model_max_length = 1024,
                 max_source_length = 1024,
                 max_target_length = 128,
                 ):

        self.checkpoint = checkpoint
        self.model_max_length = model_max_length
        self.max_source_length = max_source_length
        self.max_target_length = max_target_length
        self.model_class = model_class
        self.prompt = prompt

        if tokenizer is None:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.checkpoint,
                model_max_length = self.model_max_length,
                cache_dir=BASE_CACHE_DIR,
                use_fast=True
            )
        else:
            self.tokenizer = tokenizer


    def tokenize_and_align_labels(self, examples, inputs_column="text", labels_column="summary"):

        """

        Process the dataset for model input, for example, do tokenization and prepare label_ids.

        Args:
            examples (`Dataset`): { "text": [s1, s2, ...], "summary": [l1, l2, ...]}
            inputs (`str`): The name of input column
            labels (`str`): The name of target column

        Returns:
            `Dict`: {"input_ids": input_ids, "attention_mask": attention_mask, "labels": label_ids }

        """

        # Select input column
        inputs = examples[inputs_column]
        inputs = [self.prompt + raw_text for raw_text in inputs]
        # Setup the tokenizer for inputs
        model_inputs = self.tokenizer(inputs, max_length=self.max_source_length, padding="max_length", truncation=True)

        # Select target column
        if labels_column in examples.keys():
            labels = examples[labels_column]
            # Setup the tokenizer for targets
            with self.tokenizer.as_target_tokenizer():
                labels = self.tokenizer(labels, max_length=self.max_target_length, padding="max_length", truncation=True)
                # Ignore padding in the loss
                labels["input_ids"] = [
                    [(l if l != self.tokenizer.pad_token_id else -100) for l in label] for label in labels["input_ids"]
                ]
            model_inputs["labels"] = labels["input_ids"]

        return model_inputs

    def collator(self):

        """

        The collating function.

        Returns:
            `function`: A collating function.

            For example, **DataCollatorForSeq2Seq(...)**.

            You can also custom a collating function, but remember that `collator()` needs to return a **function**.
        """


        return DataCollatorForSeq2Seq(self.tokenizer, model=self.model_class, pad_to_multiple_of=8)

    def postprocess(self, preds, labels):

        """
        Process model's outputs and get the final results rather than simple ids.

        Args:
            preds (Tensor): Prediction labels, the output of the model.
            labels (Tensor): True labels

        Returns:
            `(LongTensor, LongTensor)`: decoded_preds, decoded_labels

        """

        decoded_preds = self.tokenizer.batch_decode(preds, skip_special_tokens=True)

        labels = np.array(labels.to("cpu"))
        # Replace -100 in the labels as we can't decode them.
        labels = np.where(labels != -100, labels, self.tokenizer.pad_token_id)

        decoded_labels = self.tokenizer.batch_decode(labels, skip_special_tokens=True)

        decoded_preds = [pred.strip() for pred in decoded_preds]
        decoded_labels = [label.strip() for label in decoded_labels]

        # rougeLSum expects newline after each sentence
        decoded_preds = ["\n".join(nltk.sent_tokenize(pred)) for pred in decoded_preds]
        decoded_labels = ["\n".join(nltk.sent_tokenize(label)) for label in decoded_labels]

        return decoded_preds, decoded_labels

    def get_dataloader(self, dataset, inputs_column="text", labels_column="summary"):

        """
        Generate DataLoader for a dataset.

        Args:
            dataset (`Dataset`): The raw dataset.
            inputs_column (`str`): Column name of the inputs.
            labels_column (`str`): Column name of the labels.

        Returns:
            `DataLoader`: A dataloader for the dataset. Will be used for inference.
        """

        tokenized_example = dataset.map(
            lambda x: self.tokenize_and_align_labels(x, inputs_column=inputs_column, labels_column=labels_column),
            batched=True,
            remove_columns=dataset.column_names
        )

        dataloader = DataLoader(
            dataset=tokenized_example,
            batch_size=8,
            collate_fn=self.collator(),
        )

        return dataloader



class DataUtilsForCTRLSum():
    """

    Args:
        tokenizer (`PretrainedTokenizer`, default to None):
            The tokenizer for tokenization.
        checkpoint (`str`):
            The checkpoint from which the tokenizer is loaded.
        model_max_length (`int`, *optional*): The max sequence length the model accepts.
        max_source_length (`int`, *optional*): The max length of the input text.
        max_target_length (`int`, *optional*): The max length of the generated summary.
    """


    def __init__(self, tokenizer = None,
                 model_class = FlanT5ForSummarization,
                 checkpoint = "google/flan-t5-base",
                 prompt = "Please give a summary of the following text ",
                 model_max_length = 1024,
                 max_source_length = 1024,
                 max_target_length = 500,
                 ):

        self.checkpoint = checkpoint
        self.model_max_length = model_max_length
        self.max_source_length = max_source_length
        self.max_target_length = max_target_length
        self.model_class = model_class
        self.prompt = prompt

        if tokenizer is None:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.checkpoint,
                model_max_length = self.model_max_length,
                cache_dir=BASE_CACHE_DIR,
                use_fast=True
            )
        else:
            self.tokenizer = tokenizer


    def tokenize_and_align_labels(self, examples, inputs_column="text", labels_column="summary"):

        """

        Process the dataset for model input, for example, do tokenization and prepare label_ids.

        Args:
            examples (`Dataset`): { "text": [s1, s2, ...], "summary": [l1, l2, ...]}
            inputs (`str`): The name of input column
            labels (`str`): The name of target column

        Returns:
            `Dict`: {"input_ids": input_ids, "attention_mask": attention_mask, "labels": label_ids }

        """

        inputs = examples[inputs_column]
        prompts = [ self.prompt for i in inputs ]

        # kw_instructions = ["{}"]
        def kw(prompt, keyword):
            # i = randint(1,10)
            return "Keywords: " + str(", ".join(keyword)) + ". " + prompt + "talking about these keywords " if keyword is not None else prompt

            # return prompt + "that focuses on " + str(", ".join(keyword)) + " " if keyword is not None else prompt
        def leng(prompt, length):
            return prompt + ", which has less than " + str(length) + " words " if length is not None else prompt

        # if "keywords" in examples.keys():
        #     keywords = examples["keywords"]
        #     if keywords is not None:
        #         prompts = [ kw(prompt,keyword) for (prompt,keyword) in zip(prompts,keywords) ]
        if "length" in examples.keys():
            if examples["length"] is not None:
                lengths = examples["length"]
                prompts = [ leng(prompt,length) for (prompt,length) in zip(prompts,lengths)]
        # kwords = ["" for i in inputs]
        # if "keywords" in examples.keys():
        #     keywords = examples["keywords"]
        #     if keywords is not None:
        #         keywords = [ keyword if keyword is not None else [] for keyword in keywords ]
        #         # print(keywords)
        #         kwords = [ " ".join(keyword) for keyword in keywords ]
        inputs = [ prompt + ": " + raw_text for (prompt,raw_text) in zip(prompts, inputs) ]
        # inputs = [ f"Which entity is this text about? {text} Entity:".format(text) for text in inputs]
        print(inputs[0][:200])



        # if "keywords" in examples.keys():
        #     keywords = examples["keywords"]
        #     if keywords is not None and keywords[0] is not None:
        #         prompts = [ prompt + "that focuses on " + str(", ".join(keyword)) + " " for (prompt,keyword) in zip(prompts,keywords) ]
        # if "length" in examples.keys():
        #     if examples["length"] is not None and examples["length"][0] is not None:
        #         lengths = [l for l in examples["length"]]
        #         prompts = [ prompt + ", which has less than " + str(length) + " words " for (prompt,length) in zip(prompts,lengths)]
        # inputs = [ prompt + ": " + raw_text for (prompt,raw_text) in zip(prompts,inputs) ]
        # print(inputs[0][:200])
        #


        # if labels_column in examples.keys():
        #     lengths = [len(s.split(" ")) for s in examples[labels_column]]
        #     lengths = [ 10*math.ceil(s/10) for s in lengths]

        # Select input column
        # inputs = examples[inputs_column]
        # inputs = [ self.prompt + "which has less than " + str(length) + " words: " + raw_text for (raw_text,length) in zip(inputs,lengths) ]
        # inputs = [ "| " + str(length) + " words | " + "summarize: " + raw_text for (raw_text, length) in
        #           zip(inputs, lengths)]

        # Setup the tokenizer for inputs

        model_inputs = self.tokenizer(inputs, max_length=self.max_source_length, padding="max_length", truncation=True)

        # Select target column
        if labels_column in examples.keys():
            labels = examples[labels_column]
            # Setup the tokenizer for targets
            with self.tokenizer.as_target_tokenizer():
                labels = self.tokenizer(labels, max_length=self.max_target_length, padding="max_length", truncation=True)
                # Ignore padding in the loss
                labels["input_ids"] = [
                    [(l if l != self.tokenizer.pad_token_id else -100) for l in label] for label in labels["input_ids"]
                ]
            model_inputs["labels"] = labels["input_ids"]


        return model_inputs

    def collator(self):

        """

        The collating function.

        Returns:
            `function`: A collating function.

            For example, **DataCollatorForSeq2Seq(...)**.

            You can also custom a collating function, but remember that `collator()` needs to return a **function**.
        """


        return DataCollatorForSeq2Seq(self.tokenizer, model=self.model_class, pad_to_multiple_of=8)

    def postprocess(self, preds, labels):

        """
        Process model's outputs and get the final results rather than simple ids.

        Args:
            preds (Tensor): Prediction labels, the output of the model.
            labels (Tensor): True labels

        Returns:
            `(LongTensor, LongTensor)`: decoded_preds, decoded_labels

        """

        decoded_preds = self.tokenizer.batch_decode(preds, skip_special_tokens=True)

        labels = np.array(labels.to("cpu"))
        # Replace -100 in the labels as we can't decode them.
        labels = np.where(labels != -100, labels, self.tokenizer.pad_token_id)

        decoded_labels = self.tokenizer.batch_decode(labels, skip_special_tokens=True)

        decoded_preds = [pred.strip() for pred in decoded_preds]
        decoded_labels = [label.strip() for label in decoded_labels]

        # rougeLSum expects newline after each sentence
        decoded_preds = ["\n".join(nltk.sent_tokenize(pred)) for pred in decoded_preds]
        decoded_labels = ["\n".join(nltk.sent_tokenize(label)) for label in decoded_labels]

        return decoded_preds, decoded_labels

    def get_dataloader(self, dataset, inputs_column="text", labels_column="summary"):

        """
        Generate DataLoader for a dataset.

        Args:
            dataset (`Dataset`): The raw dataset.
            inputs_column (`str`): Column name of the inputs.
            labels_column (`str`): Column name of the labels.

        Returns:
            `DataLoader`: A dataloader for the dataset. Will be used for inference.
        """

        tokenized_example = dataset.map(
            lambda x: self.tokenize_and_align_labels(x, inputs_column=inputs_column, labels_column=labels_column),
            batched=True,
            remove_columns=dataset.column_names
        )

        dataloader = DataLoader(
            dataset=tokenized_example,
            batch_size=8,
            collate_fn=self.collator(),
        )

        return dataloader



import argparse
import glob
import io
import json
import ntpath
import os
import time
from typing import List

from SciAssist.bin.doc2json.doc2json.grobid2json.grobid.client import ApiClient

'''
This version uses the standard ProcessPoolExecutor for parallelizing the concurrent calls to the GROBID services.
Given the limits of ThreadPoolExecutor (input stored in memory, blocking Executor.map until the whole input
is acquired), it works with batches of PDF of a size indicated in the config.json file (default is 1000 entries).
We are moving from first batch to the second one only when the first is entirely processed - which means it is
slightly sub-optimal, but should scale better. However acquiring a list of million of files in directories would
require something scalable too, which is not implemented for the moment.
'''

DEFAULT_GROBID_CONFIG = {
    "grobid_server": "localhost",
    "grobid_port": "8070",
    "batch_size": 1000,
    "sleep_time": 5,
    "generateIDs": False,
    "consolidate_header": False,
    "consolidate_citations": False,
    "include_raw_citations": True,
    "include_raw_affiliations": False,
    "max_workers": 2,
}

class GrobidClient(ApiClient):

    def __init__(self, config=None):
        self.config = config or DEFAULT_GROBID_CONFIG
        self.generate_ids = self.config["generateIDs"]
        self.consolidate_header = self.config["consolidate_header"]
        self.consolidate_citations = self.config["consolidate_citations"]
        self.include_raw_citations = self.config["include_raw_citations"]
        self.include_raw_affiliations = self.config["include_raw_affiliations"]
        self.max_workers = self.config["max_workers"]
        self.grobid_server = self.config["grobid_server"]
        self.grobid_port = self.config["grobid_port"]
        self.sleep_time = self.config["sleep_time"]

    def process(self, input: str, output: str, service: str):
        batch_size_pdf = self.config['batch_size']
        pdf_files = []

        for pdf_file in glob.glob(input + "/*.pdf"):
            pdf_files.append(pdf_file)

            if len(pdf_files) == batch_size_pdf:
                self.process_batch(pdf_files, output, service)
                pdf_files = []

        # last batch
        if len(pdf_files) > 0:
            self.process_batch(pdf_files, output, service)

    def process_batch(self, pdf_files: List[str], output: str, service: str) -> None:
        print(len(pdf_files), "PDF files to process")
        for pdf_file in pdf_files:
            self.process_pdf(pdf_file, output, service)

    def process_pdf_stream(self, pdf_file: str, pdf_strm: bytes, output: str, service: str) -> str:
        # process the stream
        files = {
            'input': (
                pdf_file,
                pdf_strm,
                'application/pdf',
                {'Expires': '0'}
            )
        }

        the_url = 'http://' + self.grobid_server
        the_url += ":" + self.grobid_port
        the_url += "/api/" + service

        # set the GROBID parameters
        the_data = {}
        if self.generate_ids:
            the_data['generateIDs'] = '1'
        else:
            the_data['generateIDs'] = '0'

        if self.consolidate_header:
            the_data['consolidateHeader'] = '1'
        else:
            the_data['consolidateHeader'] = '0'

        if self.consolidate_citations:
            the_data['consolidateCitations'] = '1'
        else:
            the_data['consolidateCitations'] = '0'

        if self.include_raw_affiliations:
            the_data['includeRawAffiliations'] = '1'
        else:
            the_data['includeRawAffiliations'] = '0'

        if self.include_raw_citations:
            the_data['includeRawCitations'] = '1'
        else:
            the_data['includeRawCitations'] = '0'

        res, status = self.post(
            url=the_url,
            files=files,
            data=the_data,
            headers={'Accept': 'text/plain'}
        )

        if status == 503:
            time.sleep(self.sleep_time)
            return self.process_pdf_stream(pdf_file, pdf_strm, service)
        elif status != 200:
            with open(os.path.join(output, "failed.log"), "a+") as failed:
                failed.write(pdf_file.strip(".pdf") + "\n")
            print('Processing failed with error ' + str(status))
            return ""
        else:
            return res.text

    def process_pdf(self, pdf_file: str, output: str, service: str) -> None:
        # check if TEI file is already produced
        # we use ntpath here to be sure it will work on Windows too
        pdf_file_name = ntpath.basename(pdf_file)
        filename = os.path.join(output, os.path.splitext(pdf_file_name)[0] + '.tei.xml')
        if os.path.isfile(filename):
            return

        print(pdf_file)
        pdf_strm = open(pdf_file, 'rb').read()
        tei_text = self.process_pdf_stream(pdf_file, pdf_strm, output, service)

        # writing TEI file
        if tei_text:
            with io.open(filename, 'w+', encoding='utf8') as tei_file:
                tei_file.write(tei_text)

    def process_citation(self, bib_string: str, log_file: str) -> str:
        # process citation raw string and return corresponding dict
        the_data = {
            'citations': bib_string,
            'consolidateCitations': '0'
        }

        the_url = 'http://' + self.grobid_server
        the_url += ":" + self.grobid_port
        the_url += "/api/processCitation"

        for _ in range(5):
            try:
                res, status = self.post(
                    url=the_url,
                    data=the_data,
                    headers={'Accept': 'text/plain'}
                )
                if status == 503:
                    time.sleep(self.sleep_time)
                    continue
                elif status != 200:
                    with open(log_file, "a+") as failed:
                        failed.write("-- BIBSTR --\n")
                        failed.write(bib_string + "\n\n")
                    break
                else:
                    return res.text
            except Exception:
                continue

    def process_header_names(self, header_string: str, log_file: str) -> str:
        # process author names from header string
        the_data = {
            'names': header_string
        }

        the_url = 'http://' + self.grobid_server
        the_url += ":" + self.grobid_port
        the_url += "/api/processHeaderNames"

        res, status = self.post(
            url=the_url,
            data=the_data,
            headers={'Accept': 'text/plain'}
        )

        if status == 503:
            time.sleep(self.sleep_time)
            return self.process_header_names(header_string, log_file)
        elif status != 200:
            with open(log_file, "a+") as failed:
                failed.write("-- AUTHOR --\n")
                failed.write(header_string + "\n\n")
        else:
            return res.text

    def process_affiliations(self, aff_string: str, log_file: str) -> str:
        # process affiliation from input string
        the_data = {
            'affiliations': aff_string
        }

        the_url = 'http://' + self.grobid_server
        the_url += ":" + self.grobid_port
        the_url += "/api/processAffiliations"

        res, status = self.post(
            url=the_url,
            data=the_data,
            headers={'Accept': 'text/plain'}
        )

        if status == 503:
            time.sleep(self.sleep_time)
            return self.process_affiliations(aff_string, log_file)
        elif status != 200:
            with open(log_file, "a+") as failed:
                failed.write("-- AFFILIATION --\n")
                failed.write(aff_string + "\n\n")
        else:
            return res.text


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Client for GROBID services")
    parser.add_argument("service", help="one of [processFulltextDocument, processHeaderDocument, processReferences]")
    parser.add_argument("--input", default=None, help="path to the directory containing PDF to process")
    parser.add_argument("--output", default=None, help="path to the directory where to put the results")
    parser.add_argument("--config", default=None, help="path to the config file, default is ./config.json")

    args = parser.parse_args()

    input_path = args.input
    config = json.load(open(args.config)) if args.config else DEFAULT_GROBID_CONFIG
    output_path = args.output
    service = args.service

    client = GrobidClient(config=config)

    start_time = time.time()

    client.process(input_path, output_path, service)

    runtime = round(time.time() - start_time, 3)
    print("runtime: %s seconds " % (runtime))



import torch
from torch import nn
from transformers import AutoModel
from transformers.modeling_outputs import TokenClassifierOutput


class BertForTokenClassifier(nn.Module):
    def __init__(
        self,
        model_checkpoint: str = "allenai/scibert_scivocab_uncased",
        output_size: int = 13,
        cache_dir: str = ".cache",
        save_name: str = "scibert_uncased.pt",
        model_dir: str = "pretrained"
    ):
        super().__init__()
        self.save_name = save_name
        self.model_dir = model_dir
        self.bert_embedder = AutoModel.from_pretrained(
            model_checkpoint,
            cache_dir=cache_dir,
        )
        print(f"The model checkpoint are cached in '{cache_dir}'.")
        self.output_size = output_size
        self.dropout = nn.Dropout(0.1)
        self.classifier = nn.Linear(in_features=768, out_features=output_size, bias=True)

    def forward(self, input_ids=None, token_type_ids=None, attention_mask=None, labels=None, token_mapping=None):

        outputs = self.bert_embedder(input_ids, attention_mask=attention_mask)
        outputs = torch.matmul(token_mapping, outputs[0])
        outputs = self.dropout(outputs)
        logits = self.classifier(outputs)

        loss = None
        if labels is not None:
            loss_fct = nn.CrossEntropyLoss()
            loss = loss_fct(logits.view(-1, self.output_size), labels.view(-1))
        return TokenClassifierOutput(
            loss=loss,
            logits=logits
        )



# main developer: Yixi Ding <dingyixi@hotmail.com>
from pathlib import Path
from typing import Optional

import datasets
from datasets import Dataset, DatasetDict
from pytorch_lightning import LightningDataModule
from torch.utils.data import DataLoader

from SciAssist.utils.data_utils import DataUtilsForSeq2Seq


class MupDataModule(LightningDataModule):
    def __init__(
        self,
        data_repo: str,
        train_batch_size: int = 8,
        num_workers: int = 0,
        pin_memory: bool = False,
        data_cache_dir: str = ".cache",
        seed: int = 777,
        data_utils = DataUtilsForSeq2Seq
    ):
        super().__init__()
        self.save_hyperparameters(logger=False)
        self.data_cache_dir = Path(self.hparams.data_cache_dir)
        self.data_utils = self.hparams.data_utils
        self.data_collator = self.data_utils.collator()

        self.data_train: Optional[Dataset] = None
        self.data_val: Optional[Dataset] = None
        self.data_test: Optional[Dataset] = None

    def prepare_data(self) -> DatasetDict:
        raw_datasets = datasets.load_dataset(
            self.hparams.data_repo,
            cache_dir=self.data_cache_dir
        )

        # Get test from csv
        # raw_datasets["test"] = csv_reader("data/mup/test-release.csv")
        return raw_datasets

    def setup(self, stage: Optional[str] = None):
        if not self.data_train and not self.data_val and not self.data_test:
            processed_datasets = self.prepare_data()
            tokenized_datasets = processed_datasets.map(
                lambda x: self.data_utils.tokenize_and_align_labels(x, inputs_column="text", labels_column="summary"),
                batched=True,
                remove_columns=processed_datasets["train"].column_names,
                load_from_cache_file=True
            )
            self.data_train = tokenized_datasets["train"]
            self.data_val = tokenized_datasets["validation"]
            # If labels are not provided, delete the column "labels"
            self.data_test = tokenized_datasets["test"].remove_columns("labels")

    def train_dataloader(self):
        return DataLoader(
            dataset=self.data_train,
            batch_size=self.hparams.train_batch_size,
            num_workers=self.hparams.num_workers,
            pin_memory=self.hparams.pin_memory,
            collate_fn=self.data_collator,
            shuffle=True,
        )

    def val_dataloader(self):
        return DataLoader(
            dataset=self.data_val,
            batch_size=self.hparams.train_batch_size,
            num_workers=self.hparams.num_workers,
            pin_memory=self.hparams.pin_memory,
            collate_fn=self.data_collator,
            shuffle=False,
        )

    def test_dataloader(self):
        return DataLoader(
            dataset=self.data_test,
            batch_size=self.hparams.train_batch_size,
            num_workers=self.hparams.num_workers,
            pin_memory=self.hparams.pin_memory,
            collate_fn=self.data_collator,
            shuffle=False,
        )




# main developer: Yixi Ding <dingyixi@hotmail.com>

import torch

from SciAssist import BASE_OUTPUT_DIR, BASE_TEMP_DIR, BASE_CACHE_DIR
from SciAssist.pipelines import TASKS, load_model


class Pipeline():

    """

    Args:
        task_name (`str`):
            The task name, which is used to load model configs.
        model_name (`str`, *optional*):
            A string, the *model name* of a pretrained model provided for this task.
        device (`str`, *optional*):
            A string, `cpu` or `gpu`.
        cache_dir (`str` or `os.PathLike`, *optional*, default to "~/.cache/sciassist"):
            Path to a directory in which a downloaded pretrained model should be
            cached if the standard cache should not be used.
        output_dir (`str` or `os.PathLike`, *optional*, default to "output/result" from current work directory):
            Path to a directory in which the predicted results files should be stored.
        temp_dir (`str` or `os.PathLike`, *optional*, default to "output/.temp" from current work directory):
            Path to a directory which holds temporary files such as `.tei.xml`.

    """

    def __init__(self, task_name: str, model_name: str = "default", device="gpu",
                 cache_dir=None, output_dir=None, temp_dir=None):

        self.device = device
        self.cache_dir = cache_dir if cache_dir is not None else BASE_CACHE_DIR
        self.output_dir = output_dir if output_dir is not None else BASE_OUTPUT_DIR
        self.temp_dir = temp_dir if temp_dir is not None else BASE_TEMP_DIR

        self.config = TASKS[task_name][model_name]
        self.model_name = model_name
        self.model = load_model(config=self.config, cache_dir=self.cache_dir, device=self.device)
        if device in ["cuda", "gpu"] and torch.cuda.is_available():
            self.device = torch.device("cuda")
            self.model.cuda()
        else:
            self.device = torch.device("cpu")

        self.data_utils = self.config["data_utils"]



import os

import nltk
from SciAssist import BASE_CACHE_DIR
from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize



# root_dir = "/home/dingyx/project/SciAssist/data/FLANT5_keyword/"
# root_dir = "/home/dingyx/project/SciAssist/data/MUP_CTRLkeyword"
#
root_dir = "/home/dingyx/project/SciAssist/data/pdfs/summary_ours/"
for dirpath,dirnames,files in os.walk(root_dir):
    file_list = files
    break

from nltk.stem.porter import *
stemmer = PorterStemmer()

suc = 0
total = 0
stop_words = set(stopwords.words('english'))
for i in file_list:
    if i[-4:]==".txt" and (int(i[:-4])) % 1 == 0:
        with open(os.path.join(root_dir,i),"r") as f:
            file = f.readlines()
            file = " ".join(file)
            # print(file)
            kw = file.strip().split(" => ")[0].strip()
            kw = word_tokenize(kw)
            kw = [stemmer.stem(k.strip()) for k in kw]

            text = file.strip().split(" => ")[1].strip()
            text = word_tokenize(text)
            text = [stemmer.stem(k.strip()) for k in text]
            kw = " ".join(kw)
            text = " ".join(text)

            # f = res(kw, text)[0]
            # f = word_tokenize(f)
            # kw = word_tokenize(kw)
            # print(f)
            # print(len(f))
            # rate = len(f)/len(kw)
            # suc += rate

            # print(kw)
            # print(text)
            if kw in text:
                suc+=1
            total+=1
        # for k in kw:
        #     # print(text)
        #     if k in stop_words:
        #         continue
        #     if k in text:
        #         suc+=1
        #     total+=1
print(suc, total)
print(suc/total)


#!/usr/bin/python
import setuptools

setuptools.setup(
    name='doc2json',
    version='0.1',
    packages=setuptools.find_packages(),
    install_requires=[
    ],
    tests_require=[
    ],
    zip_safe=False,
    test_suite='py.test',
    entry_points='',
)


# main developer: Yixi Ding <dingyixi@hotmail.com>
import json
import os
from typing import List, Tuple, Optional, Dict

from datasets import Dataset

from SciAssist import BASE_TEMP_DIR, BASE_OUTPUT_DIR
from SciAssist.pipelines.pipeline import Pipeline
from SciAssist.utils.pdf2text import process_pdf_file, get_bodytext
from SciAssist.utils.windows_pdf2text import windows_get_bodytext


class Summarization(Pipeline):
    """
    The pipeline for single document summarization.

    Args:
        model_name (`str`, *optional*):
            A string, the *model name* of a pretrained model provided for this task.
        device (`str`, *optional*):
            A string, `cpu` or `gpu`.
        cache_dir (`str` or `os.PathLike`, *optional*):
            Path to a directory in which a downloaded pretrained model should be
            cached if the standard cache should not be used.
        output_dir (`str` or `os.PathLike`, *optional*):
            Path to a directory in which the predicted results files should be stored.
        temp_dir (`str` or `os.PathLike`, *optional*):
            Path to a directory which holds temporary files such as `.tei.xml`.
        tokenizer (PreTrainedTokenizer, *optional*):
            A specific tokenizer.
        checkpoint (`str` or `os.PathLike`, *optional*):
            A checkpoint for the tokenizer. You can also specify the `checkpoint` while
            using the default tokenizer.
            Can be either:

                - A string, the *model id* of a predefined tokenizer hosted inside a model repo on huggingface.co.
                  Valid model ids can be located at the root-level, like `bert-base-uncased`, or namespaced under a
                  user or organization name, like `facebook/bart-large-cnn`.
                - A path to a *directory* containing vocabulary files required by the tokenizer, for instance saved
                  using the [`~PreTrainedTokenizer.save_pretrained`] method, e.g., `./my_model_directory/`.
                - A path or url to a single saved vocabulary file if and only if the tokenizer only requires a
                  single vocabulary file (like Bert or XLNet), e.g.: `./my_model_directory/vocab.txt`. (Not
                  applicable to all derived classes)

        model_max_length (`int`, *optional*): The max sequence length the model accepts.
        max_source_length (`int`, *optional*): The max length of the input text.
        max_target_length (`int`, *optional*): The max length of the generated summary.
    """

    def __init__(
            self, model_name: str = "default", device="gpu",
            cache_dir=None,
            output_dir=None,
            temp_dir=None,
            tokenizer=None,
            checkpoint="facebook/bart-large-cnn",
            model_max_length=1024,
            max_source_length=1024,
            max_target_length=128,
            os_name=None,
    ):
        super().__init__(task_name="single-doc-summarization", model_name=model_name, device=device,
                         cache_dir=cache_dir, output_dir=output_dir, temp_dir=temp_dir)

        self.data_utils = self.data_utils(
            tokenizer=tokenizer,
            model_class=self.config["model"],
            checkpoint=checkpoint,
            model_max_length=model_max_length,
            max_source_length=max_source_length,
            max_target_length=max_target_length
        )
        self.tokenizer = self.data_utils.tokenizer
        self.os_name = os_name if os_name != None else os.name

    def predict(
            self, input: str, type: str = "pdf",
            output_dir=None,
            temp_dir=None,
            num_beams=1,
            num_return_sequences=1,
            save_results=True,
    ):
        """

        Args:
            input (`str` or `List[str]` or `os.PathLike`):
            Can be either:

                   - A string, the reference string to be parsed.
                   - A list of strings to be parsed.
                   - A path to a *.txt* file to be summarized.
                   - A path to a *.pdf* file to be summarized, a raw scientific document without processing.
                     The pipeline will automatically extract the body text from the pdf.

            type (`str`, default to `pdf`):
                The type of input, can be either:

                    - `str` or `string`.
                    - `text`or `txt` for a .txt file.
                    - `pdf` for a pdf file. This is the default value.

            output_dir (`str` or `os.PathLike`, *optional*):
                Path to a directory in which the predicted results files should be stored.
                If not provided, it will use the `output_dir` set for the pipeline.
            temp_dir (`str` or `os.PathLike`, *optional*):
                Path to a directory which holds temporary files such as `.tei.xml`.
                If not provided, it will use the `temp_dir` set for the pipeline.
            num_beams (`int`, *optional*):
                Number of beams for beam search. 1 means no beam search.
                `num_beams` should be divisible by `num_return_sequences` for group beam search.
            num_return_sequences(`int`):
                The number of independently computed returned sequences for each element in the batch.
            save_results (`bool`, default to `True`):
                Whether to save the results in a *.json* file.
                **Note**: This is invalid when `type` is set to `str` or `string`.

        Returns:
            `Dict`: { "summary": [summary1, summary2, ...], "raw_text": raw_text }


        Examples:

             >>> from SciAssist import Summarization
             >>> pipeline = Summarization()
             >>> res = pipeline.predict('N18-3011.pdf', type="pdf", num_beams=4, num_return_sequences=2)
             >>> res["summary"]
             ['The paper proposes a method for extracting structured information from scientific documents into the literature graph. The paper describes the attributes associated with nodes and edges of different types in the graph, and describes how to extract the entities mentioned in paper text. The method is evaluated on three tasks: sequence labeling, entity linking and relation extraction. ',
             'The paper proposes a method for extracting structured information from scientific documents into the literature graph. The paper describes the attributes associated with nodes and edges of different types in the graph, and describes how to extract the entities mentioned in paper text. The method is evaluated on three tasks: sequence labeling, entity linking and relation extraction.  ']

        """

        if output_dir is None:
            output_dir = self.output_dir
        if temp_dir is None:
            temp_dir = self.temp_dir

        if type in ["str", "string"]:
            results = self._summarize_for_string(example=input, num_beams=num_beams,
                                                 num_return_sequences=num_return_sequences)
        elif type in ["txt", "text"]:
            results = self._summarize_for_text(filename=input, num_beams=num_beams,
                                               num_return_sequences=num_return_sequences)
        elif type == "pdf":
            results = self._summarize_for_pdf(filename=input, output_dir=output_dir, temp_dir=temp_dir,
                                              num_beams=num_beams, num_return_sequences=num_return_sequences)
        # Save predicted results as a text file
        if save_results and type not in ["str", "string"]:
            os.makedirs(output_dir, exist_ok=True)
            output_file = os.path.basename(input)
            with open(os.path.join(output_dir, f"{output_file[:-4]}_summ.json"), "w") as output:
                output.write(json.dumps(results) + "\n")

        return results

    def _to_device(self, batch):
        if self.model_name in ["default", "bart-cnn-on-mup"]:
            return {
                "input_ids": batch["input_ids"].to(self.device),
                "attention_mask": batch["attention_mask"].to(self.device),
            }

    def _summarize(
            self,
            examples: List[str],
            num_beams=1,
            num_return_sequences=1
    ) -> List[str]:
        """
        Summarize each text in the list.
        Args:
            examples(`List[str]`): A list of texts to be summarized
            num_beams(`int`): Number of beams for beam search. 1 means no beam search.
            num_return_sequences(`int`): The number of independently computed returned sequences for each element in the batch.
        Returns:
            `List[str]`: A list of the summarization, each item corresponds to a text in the list.
        """

        # Prepare the dataset
        dict_data = {"text": examples}
        dataset = Dataset.from_dict(dict_data)

        # Tokenize for Bart, get input_ids and attention_masks
        dataloader = self.data_utils.get_dataloader(dataset)

        results = []
        for batch in dataloader:
            batch = self._to_device(batch)

            # Get token ids of summary
            pred = self.model.generate(batch["input_ids"], batch["attention_mask"], num_beams, num_return_sequences)
            # Convert token ids to text
            decoded_preds = self.tokenizer.batch_decode(pred, skip_special_tokens=True)

            results.extend(decoded_preds)

        return results

    def _summarize_for_string(
            self,
            example: str,
            num_beams=1,
            num_return_sequences=1
    ) -> Tuple[str, str]:

        """
        Summarize a text in string format.

        Args:
            example (`str`): The string to summarize.
            num_beams (`int`): Number of beams for beam search. 1 means no beam search.
            num_return_sequences(`int`): The number of independently computed returned sequences for each element in the batch.
        Returns:
           `Tuple[str, str]`:
                Predicted summarization and source text.

        """
        res = self._summarize([example], num_beams, num_return_sequences)

        return {"summary": res, "raw_text": example}

    def _summarize_for_text(
            self,
            filename: str,
            num_beams: int = 1,
            num_return_sequences: int = 1
    ) -> Tuple[str, str]:
        """

        Summarize a document from a text file.

        Args:
            num_beams (`int`): Number of beams for beam search. 1 means no beam search.
            filename (`str`): The path to the input text file.
            num_return_sequences(`int`): The number of independently computed returned sequences for each element in the batch.

        Returns:
            `Tuple[str, str]`:
                Predicted summarization and source text.

        """

        with open(filename, "r") as f:
            examples = f.readlines()
        examples = [" ".join(examples)]
        res = self._summarize(examples, num_beams, num_return_sequences)

        return {"summary": res, "raw_text": examples[0]}

    def _summarize_for_pdf(
            self,
            filename: str,
            temp_dir: Optional[str] = BASE_TEMP_DIR,
            output_dir: Optional[str] = BASE_OUTPUT_DIR,
            num_beams: int = 1,
            num_return_sequences=1
    ) -> Dict:
        """
        Summarize a document from a PDF file.

        Args:
            filename (`str`): The path to the pdf file to summarize.
            temp_dir (`Optional[str]`): The diretorcy to save intermediate file, default to `temp/`.
            output_dir (`Optional[str]`): The diretorcy to save text file, default to `output/`.
            num_return_sequences(`int`): The number of independently computed returned sequences for each element in the batch.

        Returns:
            `Dict`:
                Predicted summarization and source text.
        """
        if self.os_name == "posix":
            # Convert PDF to JSON with doc2json.
            json_file = process_pdf_file(input_file=filename, temp_dir=temp_dir, output_dir=temp_dir)
            # Extract bodytext from pdf and save them in TEXT format.
            text_file = get_bodytext(json_file=json_file, output_dir=output_dir)
        elif self.os_name == "nt":
            text_file = windows_get_bodytext(path=filename, output_dir=output_dir)

        # Do summarization
        return self._summarize_for_text(text_file, num_beams=num_beams, num_return_sequences=num_return_sequences)



import json
import os
import re
import torch
import numpy as np
import nltk
from typing import List, Tuple, Optional, Union, Dict
from transformers import AutoTokenizer

from SciAssist import BASE_OUTPUT_DIR, BASE_TEMP_DIR
from SciAssist.pipelines.pipeline import Pipeline
from SciAssist.utils.pdf2text import process_pdf_file, get_bodytext
from SciAssist.utils.windows_pdf2text import windows_get_bodytext


class DatasetExtraction(Pipeline):
    """

    The pipeline for dataset extraction.

    Args:
        model_name (`str`, *optional*):
            A string, the *model name* of a pretrained model provided for this task.

        device (`str`, *optional*):
            A string, `cpu` or `gpu`.

        cache_dir (`str` or `os.PathLike`, *optional*):
            Path to a directory in which a downloaded pretrained model should be
            cached if the standard cache should not be used.

        output_dir (`str` or `os.PathLike`, *optional*):
            Path to a directory in which the predicted results files should be stored.

        temp_dir (`str` or `os.PathLike`, *optional*):
            Path to a directory which holds temporary files such as `.tei.xml`.

        tokenizer (PreTrainedTokenizer, *optional*):
            A specific tokenizer.

        checkpoint (`str` or `os.PathLike`, *optional*):
            A checkpoint for the tokenizer. You can also specify the `checkpoint` while
            using the default tokenizer.
            Can be either:
                - A string, the *model id* of a predefined tokenizer hosted inside a model repo on huggingface.co.
                  Valid model ids can be located at the root-level, like `bert-base-uncased`, or namespaced under a
                  user or organization name, like `allenai/scibert_scivocab_uncased`.
                - A path to a *directory* containing vocabulary files required by the tokenizer, for instance saved
                  using the [`~PreTrainedTokenizer.save_pretrained`] method, e.g., `./my_model_directory/`.
                - A path or url to a single saved vocabulary file if and only if the tokenizer only requires a
                  single vocabulary file (like Bert or XLNet), e.g.: `./my_model_directory/vocab.txt`. (Not
                  applicable to all derived classes)

        model_max_length (`int`, *optional*): The max sequence length the model accepts.

    """

    def __init__(
            self, model_name: Optional[str] = "default", device: Optional[str] = "gpu",
            cache_dir = None,
            output_dir = None,
            temp_dir = None,
            tokenizer = None,
            checkpoint = "allenai/scibert_scivocab_uncased",
            model_max_length = 128,
            os_name = None,
    ):

        super().__init__(task_name = "dataset-extraction", model_name = model_name, device = device,
                         cache_dir = cache_dir, output_dir = output_dir, temp_dir = temp_dir)

        # print('load my model')
        # self.model.load_state_dict(torch.load("/home/linxiao/SciAssist-scibert-0223/src/models/scibert_ner/2023-02-27_10-40-00/roberta-base"))
        # self.model.load_state_dict(torch.load("/home/linxiao/SciAssist-scibert-0223/src/models/scibert_ner/2023-03-24_16-57-27/roberta-base"))
        
        # print('ok')

        self.data_utils = self.data_utils(
            tokenizer = tokenizer,
            checkpoint = checkpoint,
            model_max_length = model_max_length
        )

        self.tokenizer = self.data_utils.tokenizer

        self.os_name = os_name if os_name != None else os.name

        nltk.download('punkt')

    def _to_device(self, batch):
        if self.model_name in ["default"]:
            return {
                "input_subwords": batch['input_subwords'].to(self.device),
                "input_token_start_indexs": batch['input_token_start_indexs'].to(self.device),
                "attention_mask": batch["attention_mask"].to(self.device),
            }

    def extract(
            self, input, type: str = "pdf",
            output_dir = None,
            temp_dir = None,
            save_results = True,
    ):
        """

        Args:
            input (`str` or `List[str]` or `os.PathLike`):
                Can be either:
                    - A list of strings (sentences) to be extracted dataset mentions from.
                    - A path to a *.txt* file to be extracted dataset mentions from.
                    - A path to a *.pdf* file to be extracted dataset mentions from, a raw scientific document without processing. The pipeline will automatically extract the body text from the pdf.

            type (`str`, default to `pdf`):
                The type of input, can be either:
                    - `str` or `string`.
                    - `text`or `txt` for a .txt file.
                    - `pdf` for a pdf file. This is the default value.

            output_dir (`str` or `os.PathLike`, *optional*):
                Path to a directory in which the predicted results files should be stored.
                If not provided, it will use the `output_dir` set for the pipeline.

            temp_dir (`str` or `os.PathLike`, *optional*):
                Path to a directory which holds temporary files such as `.tei.xml`.
                If not provided, it will use the `temp_dir` set for the pipeline.

            save_results (`bool`, default to `True`):
                Whether to save the results in a *.json* file.
                **Note**: This is invalid when `type` is set to `str` or `string`.

        Returns:
            `Dict`: { "dataset_mentions": [[dataset_mentions1], sentence1], [[dataset_mentions2], sentence2], ...], "text": [sentence1, sentence2, ...]}
            Please note that only positive sentences will be shown in Dict['dataset_mentions'], while Dict['text'] will show all sentences.

        Examples:
             >>> from SciAssist import DatasetExtraction
             >>> pipeline = DatasetExtraction()
             >>> res = pipeline.predict('N18-3011.pdf', type="pdf")
             >>> res["dataset_mentions"]

        """

        if output_dir is None:
            output_dir = self.output_dir
        if temp_dir is None:
            temp_dir = self.temp_dir

        if type in ["str", "string"]:
            results = self._extract_for_string(example=input)
        elif type in ["txt", "text"]:
            results = self._extract_for_text(filename=input)
        elif type == "pdf":
            results = self._extract_for_pdf(filename=input, output_dir=output_dir, temp_dir=temp_dir)

        # Save predicted results as a text file
        if save_results and type not in ["str", "string"]:
            os.makedirs(output_dir, exist_ok=True)
            output_file = os.path.basename(input)
            with open(os.path.join(output_dir, f"{output_file[:-4]}_de.json"), "w") as output:
                output.write(json.dumps(results) + "\n")

        return results


    def _extract(
            self,
            examples: List[str]
    ):
        """

        Extract dataset mentions from each sentence in the list.

        Args:
            examples(`List[str]`): A list of sentences to be extracted dataset mentions.

        Returns:
            `Dict`: { "dataset_mentions": [[dataset_mentions1], sentence1], [[dataset_mentions2], sentence2], ...], "text": [sentence1, sentence2, ...]}

        """

        results = {}

        dataset_mentions = []
        dataset_set = set()
        dataset = {'data':{'text': examples}}
        dataloader = self.data_utils.get_dataloader(dataset)

        pred_output = []
        for batch in dataloader:
            batch = self._to_device(batch)
            batch_output = self.model(**batch)
            ner_output = batch_output[0]
            ner_output = ner_output.detach().cpu().numpy()
            pred_output.extend([[idx for idx in indices] for indices in np.argmax(ner_output, axis=2)]) # pred_output: List[List[tag_idx]] num_sentences * each_sentence_length

        for i, sentence in enumerate(pred_output): # for each sentence
            entity_indexes = self.find_entity_indexes(sentence) # List of entity_index pairs: [[2,2],[5,9],[16,21]]
            if len(entity_indexes) != 0:
                tokens = examples[i].strip().split(' ')
                entities = self.find_entities(entity_indexes, tokens) # List of mentions in this sentence
                dataset_set.update(entities)
                dataset_mentions.append([entities, examples[i]]) # [[List of mentions in sentence1, sentence1], [List of mentions in sentence2, sentence2], ...]

        results['all_dataset_mentions'] = list(dataset_set)
        results['dataset_mentions'] = dataset_mentions
        results['text'] = examples

        return results


    def _extract_for_string(
            self,
            example: Union[str, List[str]]
    ):
        """

        Extract dataset mentions from each sentence in the string or list.

        Args:
            example (`Union[str, List[str]]`): The string to be extracted dataset mentions from.

        Returns:
            `Dict`: { "dataset_mentions": [[dataset_mentions1], sentence1], [[dataset_mentions2], sentence2], ...], "text": [sentence1, sentence2, ...]}

        """

        if isinstance(example, list):
            examples = example
        else:
            examples = [example]

        results = self._extract(examples)

        return results


    def _extract_for_text(
            self,
            filename: str
    ):
        """

        Extract dataset mentions from each sentence of a text file and save the result as a text file.

        Args:
            filename (`str`): The path to the text file to be extracted dataset mentions from.

        Returns:
            `Dict`: { "dataset_mentions": [[dataset_mentions1], sentence1], [[dataset_mentions2], sentence2], ...], "text": [sentence1, sentence2, ...]}

        """

        with open(filename, "r") as f:
            text = f.readlines()
            sentences = []
            for i in text:
                sentences.append(i.strip())
        f.close()

        results = self._extract(sentences)

        return results


    def _extract_for_pdf(
            self,
            filename: str,
            output_dir: Optional[str] = BASE_OUTPUT_DIR,
              temp_dir: Optional[str] = BASE_TEMP_DIR
    ):
        """

        Extract dataset mentions from each sentence of a pdf file and save the result as a text file.

        Args:
            filename (`str`): The path to the pdf file to parse.
            output_dir (`Optional[str]`): The directory to save the result file, default to `result/`.
            temp_dir (`Optional[str]`): The diretorcy to save intermediate file, default to `temp/`.

        Returns:
           `List[Dict]`:
                Tagged strings, origin tokens and labels predicted by the model.

        """
        
        if self.os_name == "posix":
            # Convert PDF to JSON with doc2json.
            json_file = process_pdf_file(input_file=filename, temp_dir=temp_dir, output_dir=temp_dir)
            # Extract bodytext from pdf and save them in TEXT format.
            text_file = get_bodytext(json_file=json_file, output_dir=output_dir)
        elif self.os_name == "nt":
            text_file = windows_get_bodytext(path=filename, output_dir=output_dir)

        with open(text_file, "r") as f:
            text = f.readlines()
            sentences = nltk.sent_tokenize(text[0])
        f.close()

        results = self._extract(sentences)

        return results


    def find_entity_indexes(self, lst):
        # {0: 'B-DATA', 1: 'I-DATA', 2: 'O'}
        indexes = []
        i = 0
        for i in range(len(lst)):
            tmp = []
            if lst[i] == 0:
                tmp = [i,i]
                j = i
                while j < len(lst):
                    if j+1 == len(lst) or lst[j+1] != 1:
                        indexes.append(tmp)
                        break                    
                    else:
                        j += 1
                        tmp = [i,j]

        return indexes


    def find_entities(self, idx_lst, tokens):
        dataset_mentions = []
        for idx_pair in idx_lst:
            if idx_pair[0] == idx_pair[1]:
                dataset_mentions.append(tokens[idx_pair[0]])
            else:
                dataset_mentions.append(' '.join(tokens[idx_pair[0]:idx_pair[1] + 1]))

        return dataset_mentions





import os
from pathlib import Path
from typing import List, Optional

import hydra
from omegaconf import DictConfig
from pytorch_lightning import (
    Callback,
    LightningDataModule,
    LightningModule,
    Trainer,
    seed_everything,
)
from pytorch_lightning.loggers import LightningLoggerBase

from SciAssist import utils

log = utils.get_logger(__name__)


def train(config: DictConfig) -> Optional[float]:
    # Set seed for random number generators in pytorch, numpy and python.random
    if config.get("seed"):
        seed_everything(config.seed, workers=True)

    # Convert relative ckpt path to absolute path if necessary
    ckpt_path = config.trainer.get("resume_from_checkpoint")
    if ckpt_path and not os.path.isabs(ckpt_path):
        config.trainer.resume_from_checkpoint = os.path.join(
            hydra.utils.get_original_cwd(), ckpt_path
        )

    model_dir = Path(config.paths.get("model_dir"))
    os.makedirs(model_dir, exist_ok=True)

    # Init lightning datamodule
    log.info(f"Instantiating datamodule <{config.datamodule._target_}>")
    datamodule: LightningDataModule = hydra.utils.instantiate(config.datamodule)

    # Init lightning model
    log.info(f"Instantiating model <{config.model._target_}>")
    model: LightningModule = hydra.utils.instantiate(config.model)

    # Init lightning callbacks
    callbacks: List[Callback] = []
    if "callbacks" in config:
        for _, cb_conf in config.callbacks.items():
            if "_target_" in cb_conf:
                log.info(f"Instantiating callback <{cb_conf._target_}>")
                callbacks.append(hydra.utils.instantiate(cb_conf))

    # Init lightning loggers
    logger: List[LightningLoggerBase] = []
    if "logger" in config:
        for _, lg_conf in config.logger.items():
            if "_target_" in lg_conf:
                log.info(f"Instantiating logger <{lg_conf._target_}>")
                logger.append(hydra.utils.instantiate(lg_conf))

    # Init lightning trainer
    log.info(f"Instantiating trainer <{config.trainer._target_}>")
    trainer: Trainer = hydra.utils.instantiate(
        config.trainer, callbacks=callbacks, logger=logger, _convert_="partial"
    )

    # Send some parameters from config to all lightning loggers
    log.info("Logging hyperparameters!")
    utils.log_hyperparameters(
        config=config,
        model=model,
        datamodule=datamodule,
        trainer=trainer,
        callbacks=callbacks,
        logger=logger,
    )

    # Train the model
    if config.get("train"):
        log.info("Starting training!")
        trainer.fit(model=model, datamodule=datamodule)

    # Get metric score for hyperparameter optimization
    optimized_metric = config.get("optimized_metric")
    if optimized_metric and optimized_metric not in trainer.callback_metrics:
        raise Exception(
            "Metric for hyperparameter optimization not found! "
            "Make sure the `optimized_metric` in `hparams_search` config is correct!"
        )
    score = trainer.callback_metrics.get(optimized_metric)

    # Test the model
    if config.get("test"):
        ckpt_path = "best"
        if not config.get("train") or config.trainer.get("fast_dev_run"):
            ckpt_path = None
        log.info("Starting testing!")
        trainer.test(model=model, datamodule=datamodule, ckpt_path=ckpt_path)

    # Make sure everything closed properly
    log.info("Finalizing!")
    utils.finish(
        config=config,
        model=model,
        datamodule=datamodule,
        trainer=trainer,
        callbacks=callbacks,
        logger=logger,
    )



    # Print path to best checkpoint
    if not config.trainer.get("fast_dev_run") and config.get("train"):
        log.info(f"Best model ckpt at {trainer.checkpoint_callback.best_model_path}")

    # Return metric score for hyperparameter optimization
    return score



import yake

import pandas as pd
import os
from tqdm import tqdm

file_list = []

root_dir = "/home/yixi/project/scisumm-corpus/data/Training-Set-2019/Task2/From-ScisummNet-2019/"
for dirpath,dirnames,files in os.walk(root_dir):
    file_list = dirnames
    break


def keyword_extraction(input_text, kw_extractor):
    # with open(os.path.join(root_dir,input_text,input_text+".txt"),"r") as f:
    with open(os.path.join(root_dir, input_text, "summary",input_text + ".scisummnet_human.txt"), "r") as f:
        input_text = f.readlines()
        input_text= " ".join(input_text[1:])
    keywords_res = kw_extractor.extract_keywords(input_text)
    keyword_str = ",".join([kw[0] for kw in keywords_res])
    return keyword_str


# In[197]:


# ### Different setting of keyword extraction

# In[841]:


language = "en"  # 文档语言
max_ngram_size = 3  # N-grams
deduplication_thresold = 0.3  # 筛选阈值,越小关键词越少
deduplication_algo = 'seqm'
windowSize = 3
numOfKeywords = 5  # 最大数量

kw_extractor = yake.KeywordExtractor(lan=language,
                                     n=max_ngram_size,
                                     dedupLim=deduplication_thresold,
                                     dedupFunc=deduplication_algo,
                                     windowsSize=windowSize,
                                     top=numOfKeywords)



scisumm = {"File":file_list}
scisumm = pd.DataFrame(scisumm)
tqdm.pandas(desc='progress bar')
scisumm['Keyword']=scisumm["File"].progress_apply(keyword_extraction,args=(kw_extractor,))

scisumm.to_csv(os.path.join(root_dir,"scisumm_keywords10_yake_bigram_src.csv"), index=False)


"""
S2ORC classes
"""

from datetime import datetime
from typing import Dict, List, Optional

from SciAssist.bin.doc2json.doc2json.config import *

CORRECT_KEYS = {
    "issn": "issue",
    "type": "type_str"
}

SKIP_KEYS = {
    'link',
    'bib_id'
}

REFERENCE_OUTPUT_KEYS = {
    'figure': {'text', 'type_str', 'uris', 'num'},
    'table': {'text', 'type_str', 'content', 'num', 'html'},
    'footnote': {'text', 'type_str', 'num'},
    'section': {'text', 'type_str', 'num', 'parent'},
    'equation': {'text', 'type_str', 'latex', 'mathml', 'num'}
}

METADATA_KEYS = {
    "title", "authors", "year", "venue", "identifiers"
}


class ReferenceEntry:
    """
    Class for representing S2ORC figure and table references

    An example json representation (values are examples, not accurate):

    {
      "FIGREF0": {
        "text": "FIG. 2. Depth profiles of...",
        "latex": null,
        "type": "figure"
      },
      "TABREF2": {
        "text": "Diversity indices of...",
        "latex": null,
        "type": "table",
        "content": "",
        "html": ""
      }
    }
    """
    def __init__(
            self,
            ref_id: str,
            text: str,
            type_str: str,
            latex: Optional[str] = None,
            mathml: Optional[str] = None,
            content: Optional[str] = None,
            html: Optional[str] = None,
            uris: Optional[List[str]] = None,
            num: Optional[str] = None,
            parent: Optional[str] = None
    ):
        self.ref_id = ref_id
        self.text = text
        self.type_str = type_str
        self.latex = latex
        self.mathml = mathml
        self.content = content
        self.html = html
        self.uris = uris
        self.num = num
        self.parent = parent

    def as_json(self):
        keep_keys = REFERENCE_OUTPUT_KEYS.get(self.type_str, None)
        if keep_keys:
            return {
                k: self.__getattribute__(k) for k in keep_keys
            }
        else:
            return {
                "text": self.text,
                "type": self.type_str,
                "latex": self.latex,
                "mathml": self.mathml,
                "content": self.content,
                "html": self.html,
                "uris": self.uris,
                "num": self.num,
                "parent": self.parent
            }


class BibliographyEntry:
    """
    Class for representing S2ORC parsed bibliography entries

    An example json representation (values are examples, not accurate):

    {
        "title": "Mobility Reports...",
        "authors": [
            {
                "first": "A",
                "middle": ["A"],
                "last": "Haija",
                "suffix": ""
            }
        ],
        "year": 2015,
        "venue": "IEEE Wireless Commun. Mag",
        "volume": "42",
        "issn": "9",
        "pages": "80--92",
        "other_ids": {
            "doi": [
                "10.1109/TWC.2014.2360196"
            ],

        }
    }

    """
    def __init__(
            self,
            bib_id: str,
            title: str,
            authors: List[Dict[str, str]],
            ref_id: Optional[str] = None,
            year: Optional[int] = None,
            venue: Optional[str] = None,
            volume: Optional[str] = None,
            issue: Optional[str] = None,
            pages: Optional[str] = None,
            other_ids: Dict[str, List] = None,
            num: Optional[int] = None,
            urls: Optional[List] = None,
            raw_text: Optional[str] = None,
            links: Optional[List] = None
    ):
        self.bib_id = bib_id
        self.ref_id = ref_id
        self.title = title
        self.authors = authors
        self.year = year
        self.venue = venue
        self.volume = volume
        self.issue = issue
        self.pages = pages
        self.other_ids = other_ids
        self.num = num
        self.urls = urls
        self.raw_text = raw_text
        self.links = links

    def as_json(self):
        return {
            "ref_id": self.ref_id,
            "title": self.title,
            "authors": self.authors,
            "year": self.year,
            "venue": self.venue,
            "volume": self.volume,
            "issue": self.issue,
            "pages": self.pages,
            "other_ids": self.other_ids,
            "num": self.num,
            "urls": self.urls,
            "raw_text": self.raw_text,
            "links": self.links
        }


class Affiliation:
    """
    Class for representing affiliation info

    Example:
        {
            "laboratory": "Key Laboratory of Urban Environment and Health",
            "institution": "Chinese Academy of Sciences",
            "location": {
              "postCode": "361021",
              "settlement": "Xiamen",
              "country": "People's Republic of China"
        }
    """
    def __init__(
            self,
            laboratory: str,
            institution: str,
            location: Dict
    ):
        self.laboratory = laboratory
        self.institution = institution
        self.location = location

    def as_json(self):
        return {
            "laboratory": self.laboratory,
            "institution": self.institution,
            "location": self.location
        }


class Author:
    """
    Class for representing paper authors

    Example:

        {
          "first": "Anyi",
          "middle": [],
          "last": "Hu",
          "suffix": "",
          "affiliation": {
            "laboratory": "Key Laboratory of Urban Environment and Health",
            "institution": "Chinese Academy of Sciences",
            "location": {
              "postCode": "361021",
              "settlement": "Xiamen",
              "country": "People's Republic of China"
            }
          },
          "email": ""
        }
    """
    def __init__(
            self,
            first: str,
            middle: List[str],
            last: str,
            suffix: str,
            affiliation: Optional[Dict] = None,
            email: Optional[str] = None
    ):
        self.first = first
        self.middle = middle
        self.last = last
        self.suffix = suffix
        self.affiliation = Affiliation(**affiliation) if affiliation else {}
        self.email = email

    def as_json(self):
        return {
            "first": self.first,
            "middle": self.middle,
            "last": self.last,
            "suffix": self.suffix,
            "affiliation": self.affiliation.as_json() if self.affiliation else {},
            "email": self.email
        }


class Metadata:
    """
    Class for representing paper metadata

    Example:
    {
      "title": "Niche Partitioning...",
      "authors": [
        {
          "first": "Anyi",
          "middle": [],
          "last": "Hu",
          "suffix": "",
          "affiliation": {
            "laboratory": "Key Laboratory of Urban Environment and Health",
            "institution": "Chinese Academy of Sciences",
            "location": {
              "postCode": "361021",
              "settlement": "Xiamen",
              "country": "People's Republic of China"
            }
          },
          "email": ""
        }
      ],
      "year": "2011-11"
    }
    """
    def __init__(
            self,
            title: str,
            authors: List[Dict],
            year: Optional[str] = None,
            venue: Optional[str] = None,
            identifiers: Optional[Dict] = {}
    ):
        self.title = title
        self.authors = [Author(**author) for author in authors]
        self.year = year
        self.venue = venue
        self.identifiers = identifiers

    def as_json(self):
        return {
            "title": self.title,
            "authors": [author.as_json() for author in self.authors],
            "year": self.year,
            "venue": self.venue,
            "identifiers": self.identifiers
        }


class Paragraph:
    """
    Class for representing a parsed paragraph from Grobid xml
    All xml tags are removed from the paragraph text, all figures, equations, and tables are replaced
    with a special token that maps to a reference identifier
    Citation mention spans and section header are extracted

    An example json representation (values are examples, not accurate):

    {
        "text": "Formal language techniques BID1 may be used to study FORMULA0 (see REF0)...",
        "mention_spans": [
            {
                "start": 27,
                "end": 31,
                "text": "[1]")
        ],
        "ref_spans": [
            {
                "start": ,
                "end": ,
                "text": "Fig. 1"
            }
        ],
        "eq_spans": [
            {
                "start": 53,
                "end": 61,
                "text": "α = 1",
                "latex": "\\alpha = 1",
                "ref_id": null
            }
        ],
        "section": "Abstract"
    }
    """
    def __init__(
            self,
            text: str,
            cite_spans: List[Dict],
            ref_spans: List[Dict],
            eq_spans: Optional[List[Dict]] = [],
            section: Optional = None,
            sec_num: Optional = None
    ):
        self.text = text
        self.cite_spans = cite_spans
        self.ref_spans = ref_spans
        self.eq_spans = eq_spans
        if type(section) == str:
            if section:
                sec_parts = section.split('::')
                section_list = [[None, sec_name] for sec_name in sec_parts]
            else:
                section_list = None
            if section_list and sec_num:
                section_list[-1][0] = sec_num
        else:
            section_list = section
        self.section = section_list

    def as_json(self):
        return {
            "text": self.text,
            "cite_spans": self.cite_spans,
            "ref_spans": self.ref_spans,
            "eq_spans": self.eq_spans,
            "section": '::'.join([sec[1] for sec in self.section]) if self.section else "",
            "sec_num": self.section[-1][0] if self.section else None
        }


class Paper:
    """
    Class for representing a parsed S2ORC paper
    """
    def __init__(
            self,
            paper_id: str,
            pdf_hash: str,
            metadata: Dict,
            abstract: List[Dict],
            body_text: List[Dict],
            back_matter: List[Dict],
            bib_entries: Dict,
            ref_entries: Dict
        ):
        self.paper_id = paper_id
        self.pdf_hash = pdf_hash
        self.metadata = Metadata(**metadata)
        self.abstract = [Paragraph(**para) for para in abstract]
        self.body_text = [Paragraph(**para) for para in body_text]
        self.back_matter = [Paragraph(**para) for para in back_matter]
        self.bib_entries = [
            BibliographyEntry(
                bib_id=key,
                **{CORRECT_KEYS[k] if k in CORRECT_KEYS else k: v for k, v in bib.items() if k not in SKIP_KEYS}
            ) for key, bib in bib_entries.items()
        ]
        self.ref_entries = [
            ReferenceEntry(
                ref_id=key,
                **{CORRECT_KEYS[k] if k in CORRECT_KEYS else k: v for k, v in ref.items() if k != 'ref_id'}
            ) for key, ref in ref_entries.items()
        ]

    def as_json(self):
        return {
            "paper_id": self.paper_id,
            "pdf_hash": self.pdf_hash,
            "metadata": self.metadata.as_json(),
            "abstract": [para.as_json() for para in self.abstract],
            "body_text": [para.as_json() for para in self.body_text],
            "back_matter": [para.as_json() for para in self.back_matter],
            "bib_entries": {bib.bib_id: bib.as_json() for bib in self.bib_entries},
            "ref_entries": {ref.ref_id: ref.as_json() for ref in self.ref_entries}
        }

    @property
    def raw_abstract_text(self) -> str:
        """
        Get all the body text joined by a newline
        :return:
        """
        return '\n'.join([para.text for para in self.abstract])

    @property
    def raw_body_text(self) -> str:
        """
        Get all the body text joined by a newline
        :return:
        """
        return '\n'.join([para.text for para in self.body_text])

    def release_json(self, doc_type: str="pdf"):
        """
        Return in release JSON format
        :return:
        """
        # TODO: not fully implemented; metadata format is not right; extra keys in some places
        release_dict = {"paper_id": self.paper_id}
        release_dict.update({"header": {
            "generated_with": f'{S2ORC_NAME_STRING} {S2ORC_VERSION_STRING}',
            "date_generated": datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        }})
        release_dict.update(self.metadata.as_json())
        release_dict.update({"abstract": self.raw_abstract_text})
        release_dict.update({
            f"{doc_type}_parse": {
                "paper_id": self.paper_id,
                "_pdf_hash": self.pdf_hash,
                "abstract": [para.as_json() for para in self.abstract],
                "body_text": [para.as_json() for para in self.body_text],
                "back_matter": [para.as_json() for para in self.back_matter],
                "bib_entries": {bib.bib_id: bib.as_json() for bib in self.bib_entries},
                "ref_entries": {ref.ref_id: ref.as_json() for ref in self.ref_entries}
            }
        })
        return release_dict


def load_s2orc(paper_dict: Dict) -> Paper:
    """
    Load release S2ORC into Paper class
    :param paper_dict:
    :return:
    """
    paper_id = paper_dict['paper_id']
    pdf_hash = paper_dict.get('_pdf_hash', paper_dict.get('s2_pdf_hash', None))

    # 2019 gorc parses
    if "grobid_parse" in paper_dict and paper_dict.get("grobid_parse"):
        metadata = {k: v for k, v in paper_dict["metadata"].items() if k in METADATA_KEYS}
        abstract = paper_dict.get("grobid_parse").get("abstract", [])
        body_text = paper_dict.get("grobid_parse").get("body_text", [])
        back_matter = paper_dict.get("grobid_parse").get("back_matter", [])
        bib_entries = paper_dict.get("grobid_parse").get("bib_entries", {})
        for k, v in bib_entries.items():
            if 'link' in v:
                v['links'] = [v['link']]
        ref_entries = paper_dict.get("grobid_parse").get("ref_entries", {})
    # current and 2020 s2orc release_json
    elif ("pdf_parse" in paper_dict and paper_dict.get("pdf_parse")) or ("body_text" in paper_dict and paper_dict.get("body_text")):
        if "pdf_parse" in paper_dict:
            paper_dict = paper_dict["pdf_parse"]
        if paper_dict.get("metadata"):
            metadata = {k: v for k, v in paper_dict.get("metadata").items() if k in METADATA_KEYS}
        # 2020 s2orc releases (metadata is separate)
        else:
            metadata = {
                "title": None,
                "authors": [],
                "year": None
            }
        abstract = paper_dict.get("abstract", [])
        body_text = paper_dict.get("body_text", [])
        back_matter = paper_dict.get("back_matter", [])
        bib_entries = paper_dict.get("bib_entries", {})
        for k, v in bib_entries.items():
            if 'link' in v:
                v['links'] = [v['link']]
        ref_entries = paper_dict.get("ref_entries", {})
    else:
        print(paper_id)
        raise NotImplementedError("Unknown S2ORC file type!")

    return Paper(
        paper_id=paper_id,
        pdf_hash=pdf_hash,
        metadata=metadata,
        abstract=abstract,
        body_text=body_text,
        back_matter=back_matter,
        bib_entries=bib_entries,
        ref_entries=ref_entries
    )


import os

from pdfminer.high_level import extract_pages

from SciAssist import BASE_OUTPUT_DIR

path="test.pdf"
def process_pdf(path):
    raw_text = []
    res = {
        "title":"",
        "author":[],
        "body_text":[],
        "reference":[]
    }
    print(path)
    fp = open(path, 'rb')
    for page_layout in extract_pages(path):
        for element in page_layout:
            if "get_text" in dir(element):
                text = element.get_text().replace("\n","")
                text = text.strip()
                if text.isdigit() == False and text != "":
                    raw_text.append(text)
    fp.close()
    res["title"] = raw_text[0]

    i = 1 # the index of the current text in raw_text

    while i<len(raw_text):
        if raw_text[i] in ["abstract","Abstract", "ABSTRACT"]:
            i += 1
            break
        res["author"].append(raw_text[i])
        i += 1

    while i<len(raw_text):
        if "Introduction" in raw_text[i] or "INTRODUCTION" in raw_text[i]:
            break
        i += 1

    while i<len(raw_text):
        if raw_text[i] in ["References","Reference", "REFERENCE", "REFERENCES"]:
            i += 1
            break
        res["body_text"].append(raw_text[i])
        i += 1

    while i<len(raw_text):
        if "Appendix" in raw_text[i]:
            i += 1
            break
        res["reference"].append(raw_text[i])
        i += 1

    return res


def windows_get_reference(path, output_dir: str = BASE_OUTPUT_DIR):

    os.makedirs(output_dir, exist_ok=True)

    res = process_pdf(path)
    output_path = os.path.join(output_dir, os.path.basename(path)[:-4] + "_ref.txt")
    with open(output_path,"w",encoding="utf-8") as output:
        for i in res["reference"]:
            output.write(i+"\n")
    return output_path

def windows_get_bodytext(path, output_dir: str = BASE_OUTPUT_DIR):

    os.makedirs(output_dir, exist_ok=True)

    res = process_pdf(path)
    output_path = os.path.join(output_dir, os.path.basename(path)[:-4] + "_body.txt")
    with open(output_path,"w",encoding="utf-8") as output:
        for i in res["body_text"]:
            output.write(i+" ")
    return output_path



# main developer: Yixi Ding <dingyixi@hotmail.com>
import csv
import math
import os
from pathlib import Path
from typing import Optional

import datasets
from datasets import Dataset, DatasetDict
from pytorch_lightning import LightningDataModule
from torch.utils.data import DataLoader, random_split

from SciAssist.utils.data_reader import csv_reader
from SciAssist.utils.data_utils import DataUtilsForSeq2Seq
import nltk
from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize



class TestDataModule(LightningDataModule):
    def __init__(
        self,
        data_repo: str,
        train_batch_size: int = 8,
        num_workers: int = 0,
        pin_memory: bool = False,
        data_cache_dir: str = ".cache",
        seed: int = 777,
        data_utils = DataUtilsForSeq2Seq
    ):
        super().__init__()
        self.save_hyperparameters(logger=False)
        self.data_cache_dir = Path(self.hparams.data_cache_dir)
        self.data_utils = self.hparams.data_utils
        self.data_collator = self.data_utils.collator()

        self.data_test: Optional[Dataset] = None

    def prepare_data(self) -> DatasetDict:

        stop_words = set(stopwords.words('english'))
        # Prepare keywords
        text = []
        summ = []
        kw = []
        id = []
        lengths = []
        with open("/home/dingyx/project/SciAssist/data/MUP/mup_flant5_target_Entities_sci2.csv", 'r', newline='', encoding='ISO-8859-1') as f:
            rows = csv.reader(f)
            # Get Column names
            keys = next(rows)
            # Add values by column
            for row in rows:
                # keywords = list(set(row[4].split(",")))
                # kws = []
                # for k in keywords:
                #     word_tokens = word_tokenize(k)
                #     for w in word_tokens:
                #         if w not in stop_words:
                #             kws.append(w)


                kws = row[4].split(",")

                kws = [kws[0]]
                # if int(row[0]) <= 300 and int(row[0]) % 10 == 0:
                #     with open("/home/dingyx/project/SciAssist/data/Flan-T5/" + row[0] +".txt","w") as f:
                #         f.write(row[3])
                #         f.write("\n\n")
                #         f.write(" ".join(kws))
                #         f.write(" => ")
                #     text.append(row[3])
                #     summ.append(row[2])
                #     kw.append(kws)
                #     id.append(int(row[0]))

                # with open("/home/dingyx/project/SciAssist/data/MUP_all2/" + row[0] + ".txt", "w") as f:
                #     f.write(" ".join(kws) + " => ")
                for i in range(5):
                    text.append(row[3])
                    summ.append(row[2])
                    kw.append(kws)
                    id.append(int(row[0]))
                    lengths.append((i+1)*50)


        # lengths = [len(s.split(" ")) for s in summ]
        # lengths = [50*math.ceil(s/50) for s in lengths]
        # lengths = [50 for s in lengths]
        mup_datasets = {
            "text": text,
            "summary": summ,
            # "keywords": kw,
            "length": lengths,
            "id": id,
            # "length": [None for i in text]
        }
        # raw_datasets["validation"] = {


        raw_datasets = DatasetDict()
        raw_datasets["test"] = Dataset.from_dict(mup_datasets)
        # raw_datasets["validation"] = Dataset.from_dict(raw_datasets["validation"])
        # raw_datasets = Dataset.from_dict(raw_datasets)
        return raw_datasets


    def setup(self, stage: Optional[str] = None):
        if not self.data_test:
            processed_datasets = self.prepare_data()
            tokenized_datasets = processed_datasets.map(
                lambda x: self.data_utils.tokenize_and_align_labels(x, inputs_column="text", labels_column="summary"),
                batched=True,
                remove_columns=processed_datasets["test"].column_names,
                load_from_cache_file=True
            )
            # self.data_train = tokenized_datasets["train"]
            # self.data_val = tokenized_datasets["validation"]
            self.data_test = tokenized_datasets["test"]
            # If labels are not provided, delete the column "labels"
            # self.data_test = tokenized_datasets["test"]



    def test_dataloader(self):
        return DataLoader(
            dataset=self.data_test,
            batch_size=self.hparams.train_batch_size,
            num_workers=self.hparams.num_workers,
            pin_memory=self.hparams.pin_memory,
            collate_fn=self.data_collator,
            shuffle=False,
        )




import os
import platform

ROOT_DIR = os.getcwd()
BASE_OUTPUT_DIR = os.path.join(ROOT_DIR, "output/result")
BASE_TEMP_DIR = os.path.join(ROOT_DIR,"output/.temp")
os.environ["TOKENIZERS_PARALLELISM"] = "false"

USER_DIR_MAP = {
    "linux": "HOME",
    "darwin": "HOME",
    "windows": "USERPROFILE"
}
USER_DIR = os.environ[USER_DIR_MAP[platform.system().lower()]]
# Directory for cached files, including checkpoints, model dicts, tokenizers involved in pipelines.
# Unlike `cache_dir` set in config files, this is for users.

BASE_CACHE_DIR = os.path.join(USER_DIR, ".cache/sciassist")


from SciAssist.pipelines.reference_string_parsing import ReferenceStringParsing
from SciAssist.pipelines.summarization import Summarization
from SciAssist.pipelines.dataset_extraction import DatasetExtraction


import os
from pathlib import Path
from typing import Any, List

import matplotlib.pyplot as plt
import seaborn as sn
import torch
from pytorch_lightning import LightningModule
import sklearn

from SciAssist.models.components.bert_dataset_extraction import BertForDatasetExtraction
from SciAssist.utils.data_utils import DataUtilsForDatasetExtraction
from seqeval.metrics import f1_score, precision_score, recall_score, classification_report
from seqeval.scheme import IOB2
from sklearn.metrics import precision_score as sc_precision_score, recall_score as sc_recall_score, f1_score as sc_f1_score

class DatasetExtractionModule(LightningModule):
    def __init__(
        self,
        model: BertForDatasetExtraction,
        data_utils: DataUtilsForDatasetExtraction,
        lr: float = 1e-5,
    ):
        super().__init__()
        self.save_hyperparameters(logger=False)
        self.data_utils = data_utils
        self.model = model

        self.val_f1 = 0
        self.test_f1 = 0
        self.val_pre = 0
        self.test_pre = 0
        self.val_recall = 0
        self.test_recall = 0

        self.val_cls_f1 = 0
        self.test_cls_f1 = 0
        self.val_cls_pre = 0
        self.test_cls_pre = 0
        self.val_cls_recall = 0
        self.test_cls_recall = 0

        self.val_f1_best = 0

        self.save_path = Path(os.path.join(self.model.model_dir,self.model.save_name))


    def forward(self, inputs):
        return self.hparams.model(**inputs)


    def on_train_start(self):
        self.val_acc_best = 0
        self.val_f1_best = 0


    def step(self, batch: Any):
        inputs, batch_tags, batch_labels = batch, batch["ner_tags"], batch["cls_labels"]
        outputs = self.forward(inputs)

        loss = outputs[0]
        ner_output = outputs[1]
        cls_output = outputs[2]

        return loss, ner_output, cls_output, batch_tags, batch_labels


    def training_step(self, batch: Any, batch_idx: int):
        loss, ner_output, cls_output, batch_tags, batch_labels = self.step(batch)
        self.log("train/loss", loss, on_step=False, on_epoch=True, prog_bar=False)
        return {"loss": loss}


    def training_epoch_end(self, outputs: List[Any]):
        pass


    def validation_step(self, batch: Any, batch_idx: int):
        loss, ner_output, cls_output, batch_tags, batch_labels = self.step(batch)
        pred_tags, true_tags, pred_labels, true_labels = self.data_utils.postprocess(ner_output, cls_output, batch_tags, batch_labels)

        self.val_f1 = f1_score(true_tags, pred_tags, mode='strict', scheme=IOB2)
        self.val_cls_f1 = sc_f1_score(pred_labels, true_labels, average='binary')

        self.log("val/loss", loss, on_step=False, on_epoch=True, prog_bar=False)
        self.log("val/f1", self.val_f1, on_step=False, on_epoch=True, prog_bar=False)
        self.log("val/cls_f1", self.val_cls_f1, on_step=False, on_epoch=True, prog_bar=False)
        return {"loss": loss, "pred_tags": pred_tags, "true_tags": true_tags, "pred_labels": pred_labels, "true_labels": true_labels}


    def validation_epoch_end(self, outputs):
        true_tags = []
        pred_tags = []
        pred_labels = []
        true_labels = []
        for output in outputs:
            pred_tags.extend(output["pred_tags"])
            true_tags.extend(output["true_tags"])
            pred_labels.extend(output["pred_labels"])
            true_labels.extend(output["true_labels"])

        self.val_f1 = f1_score(true_tags, pred_tags, mode='strict', scheme=IOB2)
        self.val_pre = precision_score(true_tags, pred_tags, mode='strict', scheme=IOB2)
        self.val_recall = recall_score(true_tags, pred_tags, mode='strict', scheme=IOB2)
        self.val_cls_f1 = sc_f1_score(true_labels, pred_labels,average='binary')
        self.val_cls_pre = sc_precision_score(true_labels, pred_labels, average='binary')
        self.val_cls_recall = sc_recall_score(true_labels, pred_labels, average='binary')

        self.log("val/f1", self.val_f1, on_epoch=True, prog_bar=True)
        self.log("val/pre", self.val_pre, on_epoch=True, prog_bar=True)
        self.log("val/recall", self.val_recall, on_epoch=True, prog_bar=True)
        self.log("val/cls_f1", self.val_cls_f1, on_epoch=True, prog_bar=True)
        self.log("val/cls_pre", self.val_cls_pre, on_epoch=True, prog_bar=True)
        self.log("val/cls_recall", self.val_cls_recall, on_epoch=True, prog_bar=True)

        if self.val_f1 > self.val_f1_best:
            print(f"The f1 on validation set: {self.val_f1:.3f}, higher than the previous best one {self.val_f1_best:.3f}")
            print(f"Epoch: {self.current_epoch}: Found better model! Save the current model as the best model!")
            self.val_f1_best = self.val_f1
            torch.save(self.model.state_dict(), self.save_path)
        else:
            print(f"Epoch: {self.current_epoch}: No improvement! The current best model is still the best.")
        self.log("val/f1_best", self.val_f1_best, on_epoch=True, prog_bar=True)
        print("Epoch:", self.current_epoch, ", val/f1:", self.val_f1, ", val/precision:", self.val_pre, ", val/recall:", self.val_recall, ", val/cls_f1:", self.val_cls_f1, ", val/cls_pre:", self.val_cls_pre, ", val/cls_recall", self.val_cls_recall, ", val/best_f1:", self.val_f1_best)
        print(classification_report(true_tags, pred_tags, mode='strict', scheme=IOB2))


    def test_step(self, batch: Any, batch_idx: int):
        loss, ner_output, cls_output, batch_tags, batch_labels = self.step(batch)
        pred_tags, true_tags, pred_labels, true_labels = self.data_utils.postprocess(ner_output, cls_output, batch_tags, batch_labels)
        self.test_f1 = f1_score(true_tags, pred_tags, mode='strict', scheme=IOB2)
        self.test_cls_f1 = sc_f1_score(pred_labels, true_labels, average='binary')

        self.log("test/loss", loss, on_step=False, on_epoch=True, prog_bar=False)
        self.log("test/f1", self.test_f1, on_step=False, on_epoch=True, prog_bar=False)
        self.log("test/cls_f1", self.test_cls_f1, on_step=False, on_epoch=True, prog_bar=False)

        return {"loss": loss, "pred_tags": pred_tags, "true_tags": true_tags, "pred_labels": pred_labels, "true_labels": true_labels}


    def test_epoch_end(self, outputs):
        true_tags = []
        pred_tags = []
        pred_labels = []
        true_labels = []
        for output in outputs:
            pred_tags.extend(output["pred_tags"])
            true_tags.extend(output["true_tags"])
            pred_labels.extend(output["pred_labels"])
            true_labels.extend(output["true_labels"])

        self.test_f1 = f1_score(true_tags, pred_tags, mode='strict', scheme=IOB2)
        self.test_pre = precision_score(true_tags, pred_tags, mode='strict', scheme=IOB2)
        self.test_recall = recall_score(true_tags, pred_tags, mode='strict', scheme=IOB2)
        self.test_cls_f1 = sc_f1_score(true_labels, pred_labels,average='binary')
        self.test_cls_pre = sc_precision_score(true_labels, pred_labels, average='binary')
        self.test_cls_recall = sc_recall_score(true_labels, pred_labels, average='binary')

        self.log("test/f1", self.test_f1, on_epoch=True, prog_bar=True)
        self.log("test/pre", self.test_pre, on_epoch=True, prog_bar=True)
        self.log("test/recall", self.test_recall, on_epoch=True, prog_bar=True)
        self.log("test/cls_f1", self.test_cls_f1, on_epoch=True, prog_bar=True)
        self.log("test/cls_pre", self.test_cls_pre, on_epoch=True, prog_bar=True)
        self.log("test/cls_recall", self.test_cls_recall, on_epoch=True, prog_bar=True)
        print("test/f1:", self.test_f1, ", test/precision:", self.test_pre, ", test/recall:", self.test_recall, ", test/cls_f1:", self.test_cls_f1, ", test/cls_pre:", self.test_cls_pre, ", val/test_recall", self.test_cls_recall)
        print(classification_report(true_tags, pred_tags, mode='strict', scheme=IOB2))


    def on_epoch_end(self):
        self.val_f1 = 0
        self.test_f1 = 0
        self.val_pre = 0
        self.test_pre = 0
        self.val_recall = 0
        self.test_recall = 0

        self.val_cls_f1 = 0
        self.test_cls_f1 = 0
        self.val_cls_pre = 0
        self.test_cls_pre = 0
        self.val_cls_recall = 0
        self.test_cls_recall = 0


    def configure_optimizers(self):
        return torch.optim.AdamW(
            params=self.hparams.model.parameters(), lr=self.hparams.lr
        )




import argparse
import json
import os
import time
from typing import Optional, Dict

from bs4 import BeautifulSoup

from SciAssist.bin.doc2json.doc2json.grobid2json.grobid.grobid_client import GrobidClient
from SciAssist.bin.doc2json.doc2json.grobid2json.tei_to_json import convert_tei_xml_file_to_s2orc_json, \
    convert_tei_xml_soup_to_s2orc_json

BASE_TEMP_DIR = 'temp/tei_xml'
BASE_OUTPUT_DIR = 'temp/pdf'
BASE_LOG_DIR = 'log'


def process_pdf_stream(input_file: str, sha: str, input_stream: bytes, grobid_config: Optional[Dict] = None) -> Dict:
    """
    Process PDF stream
    :param input_file:
    :param sha:
    :param input_stream:
    :return:
    """
    # process PDF through Grobid -> TEI.XML
    client = GrobidClient(grobid_config)
    tei_text = client.process_pdf_stream(input_file, input_stream, 'temp', "processFulltextDocument")

    # make soup
    soup = BeautifulSoup(tei_text, "xml")

    # get paper
    paper = convert_tei_xml_soup_to_s2orc_json(soup, input_file, sha)

    return paper.release_json('pdf')


def process_pdf_file(
        input_file: str,
        temp_dir: str = BASE_TEMP_DIR,
        output_dir: str = BASE_OUTPUT_DIR,
        grobid_config: Optional[Dict] = None
) -> str:
    """
    Process a PDF file and get JSON representation
    :param input_file:
    :param temp_dir:
    :param output_dir:
    :return:
    """
    os.makedirs(temp_dir, exist_ok=True)
    os.makedirs(output_dir, exist_ok=True)

    # get paper id as the name of the file
    paper_id = '.'.join(input_file.split('/')[-1].split('.')[:-1])
    tei_file = os.path.join(temp_dir, f'{paper_id}.tei.xml')
    output_file = os.path.join(output_dir, f'{paper_id}.json')

    # check if input file exists and output file doesn't
    if not os.path.exists(input_file):
        raise FileNotFoundError(f"{input_file} doesn't exist")
    if os.path.exists(output_file):
        print(f'{output_file} already exists!')

    # process PDF through Grobid -> TEI.XML
    client = GrobidClient(grobid_config)
    # TODO: compute PDF hash
    # TODO: add grobid version number to output
    client.process_pdf(input_file, temp_dir, "processFulltextDocument")

    # process TEI.XML -> JSON
    assert os.path.exists(tei_file)
    paper = convert_tei_xml_file_to_s2orc_json(tei_file)

    # write to file
    with open(output_file, 'w') as outf:
        # print(paper.release_json())
        json.dump(paper.release_json(), outf, indent=4, sort_keys=False)
    hidden_file = os.path.join(temp_dir, f'.{paper_id}.tei.xml')
    os.rename(tei_file, hidden_file)
    return output_file


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Run S2ORC PDF2JSON")
    parser.add_argument("-i", "--input", default=None, help="path to the input PDF file")
    parser.add_argument("-t", "--temp", default=BASE_TEMP_DIR, help="path to the temp dir for putting tei xml files")
    parser.add_argument("-o", "--output", default=BASE_OUTPUT_DIR, help="path to the output dir for putting json files")
    parser.add_argument("-k", "--keep", action='store_true')

    args = parser.parse_args()

    input_path = args.input
    temp_path = args.temp
    output_path = args.output
    keep_temp = args.keep

    start_time = time.time()

    os.makedirs(temp_path, exist_ok=True)
    os.makedirs(output_path, exist_ok=True)

    process_pdf_file(input_path, temp_path, output_path)

    runtime = round(time.time() - start_time, 3)
    print("runtime: %s seconds " % (runtime))
    print('done.')